from typing import Optional

from nautilus_trader.config import TradingNodeConfig as NautilusTradingNodeConfig

from nacre.config.actors import ExposerConfig
from nacre.config.actors import PubSubConfig
from nacre.config.live import LiveExecEngineConfig


class TradingNodeConfig(NautilusTradingNodeConfig):
    """
    Configuration for ``TradingNode`` instances.

    pubsub: PubSubConfig, optional
        The config for external msgbus pubsub
    exposer: ExposerConfig, optional
        The config for exposer
    """

    exec_engine: LiveExecEngineConfig = (
        LiveExecEngineConfig()
    )  # Override default execute engine config
    pubsub: Optional[PubSubConfig] = None
    exposer: Optional[ExposerConfig] = None



# flake8: noqa
#!/usr/bin/env python3

import itertools
import os
import platform
import shutil
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import List

import numpy as np
from Cython.Build import build_ext
from Cython.Build import cythonize
from Cython.Compiler import Options
from Cython.Compiler.Version import version as cython_compiler_version
from setuptools import Distribution
from setuptools import Extension


# The Cargo rustc mode
CARGO_MODE = os.getenv("CARGO_MODE", "release")  # Release mode by default until there's an issue
# If DEBUG mode is enabled, include traces necessary for coverage, profiling and skip optimizations
DEBUG_MODE = bool(os.getenv("DEBUG_MODE", ""))
# If ANNOTATION mode is enabled, generate an annotated HTML version of the input source files
ANNOTATION_MODE = bool(os.getenv("ANNOTATION_MODE", ""))
# If PARALLEL build is enabled, uses all CPUs for compile stage of build
PARALLEL_BUILD = True if os.getenv("PARALLEL_BUILD", "true") == "true" else False
# If SKIP_BUILD_COPY is enabled, prevents copying built *.so files back into the source tree
SKIP_BUILD_COPY = bool(os.getenv("SKIP_BUILD_COPY", ""))


################################################################################
#  RUST BUILD
################################################################################
if platform.system() == "Windows":
    # Use clang as the default compiler
    os.environ["CC"] = "clang"
    os.environ["LDSHARED"] = "clang -shared"
    # https://docs.microsoft.com/en-US/cpp/error-messages/tool-errors/linker-tools-error-lnk1181?view=msvc-170&viewFallbackFrom=vs-2019
    target_dir = os.path.join(os.getcwd(), "nautilus_core", "target", "release")
    os.environ["LIBPATH"] = os.environ.get("LIBPATH", "") + f":{target_dir}"
    RUST_LIB_PFX = ""
    RUST_LIB_EXT = "lib"
    TARGET_DIR = "x86_64-pc-windows-msvc/"
else:
    RUST_LIB_PFX = "lib"
    RUST_LIB_EXT = "a"
    TARGET_DIR = ""

# Directories with headers to include
# RUST_INCLUDES = []

RUST_LIB_DIR = "debug" if CARGO_MODE in ("", "debug") else "release"

RUST_LIBS = [
    f"nautilus_core/target/{TARGET_DIR}{RUST_LIB_DIR}/{RUST_LIB_PFX}nautilus_common.{RUST_LIB_EXT}",
    f"nautilus_core/target/{TARGET_DIR}{RUST_LIB_DIR}/{RUST_LIB_PFX}nautilus_core.{RUST_LIB_EXT}",
    f"nautilus_core/target/{TARGET_DIR}{RUST_LIB_DIR}/{RUST_LIB_PFX}nautilus_model.{RUST_LIB_EXT}",
]


def _build_rust_libs() -> None:
    extra_flags = ""
    if platform.system() == "Windows":
        extra_flags = " --target x86_64-pc-windows-msvc"

    build_option = " --release" if CARGO_MODE == "release" else ""
    # Build the Rust libraries using Cargo
    print("Compiling Rust libraries...")
    build_cmd = f"(cd nautilus_core && cargo build{build_option}{extra_flags})"
    print(build_cmd)
    os.system(build_cmd)  # noqa


# Later we can be more selective about which libs are included where - to optimize binary sizes

##########################
#  Cython build options  #
##########################
# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html

Options.docstrings = True  # Include docstrings in modules
Options.fast_fail = True  # Abort the compilation on the first error occurred
Options.emit_code_comments = True
Options.annotate = ANNOTATION_MODE  # Create annotated HTML files for each .pyx
if ANNOTATION_MODE:
    Options.annotate_coverage_xml = "coverage.xml"
Options.fast_fail = True  # Abort compilation on first error
Options.warning_errors = True  # Treat compiler warnings as errors
Options.extra_warnings = True

CYTHON_COMPILER_DIRECTIVES = {
    "language_level": "3",
    "cdivision": True,  # If division is as per C with no check for zero (35% speed up)
    "embedsignature": True,  # If docstrings should be embedded into C signatures
    "profile": DEBUG_MODE,  # If we're debugging, turn on profiling
    "linetrace": DEBUG_MODE,  # If we're debugging, turn on line tracing
    "warn.maybe_uninitialized": True,
}


def _build_extensions() -> List[Extension]:
    # Regarding the compiler warning: #warning "Using deprecated NumPy API,
    # disable it with " "#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION"
    # https://stackoverflow.com/questions/52749662/using-deprecated-numpy-api
    # From the Cython docs: "For the time being, it is just a warning that you can ignore."
    define_macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
    if DEBUG_MODE or ANNOTATION_MODE:
        # Profiling requires special macro directives
        define_macros.append(("CYTHON_TRACE", "1"))

    extra_compile_args = []
    if not DEBUG_MODE and platform.system() != "Windows":
        extra_compile_args.append("-O3")
        extra_compile_args.append("-pipe")

    extra_link_args = RUST_LIBS
    # if platform.system() == "Windows":
    #     extra_link_args += [
    #         "WS2_32.Lib",
    #         "AdvAPI32.Lib",
    #         "UserEnv.Lib",
    #         "bcrypt.lib",
    #     ]

    print("Creating C extension modules...")
    print(f"define_macros={define_macros}")
    print(f"extra_compile_args={extra_compile_args}")

    return [
        Extension(
            name=str(pyx.relative_to(".")).replace(os.path.sep, ".")[:-4],
            sources=[str(pyx)],
            include_dirs=[".", np.get_include()],
            define_macros=define_macros,
            language="c",
            extra_link_args=extra_link_args,
            extra_compile_args=extra_compile_args,
        )
        for pyx in itertools.chain(Path("nacre").rglob("*.pyx"))
    ]


def _build_distribution(extensions: List[Extension]) -> Distribution:
    # Build a Distribution using cythonize()
    # Determine the build output directory
    if DEBUG_MODE:
        # For subsequent debugging, the C source needs to be in
        # the same tree as the Cython code (not in a separate build directory).
        build_dir = None
    elif ANNOTATION_MODE:
        build_dir = "build/annotated"
    else:
        build_dir = "build/optimized"

    distribution = Distribution(
        dict(
            name="nacre",
            ext_modules=cythonize(
                module_list=extensions,
                compiler_directives=CYTHON_COMPILER_DIRECTIVES,
                nthreads=os.cpu_count(),
                build_dir=build_dir,
                gdb_debug=DEBUG_MODE,
            ),
            zip_safe=False,
        )
    )
    distribution.package_dir = "nacre"
    return distribution


def _copy_build_dir_to_project(cmd: build_ext) -> None:
    # Copy built extensions back to the project tree
    for output in cmd.get_outputs():
        relative_extension = os.path.relpath(output, cmd.build_lib)
        if not os.path.exists(output):
            continue

        # Copy the file and set permissions
        shutil.copyfile(output, relative_extension)
        mode = os.stat(relative_extension).st_mode
        mode |= (mode & 0o444) >> 2
        os.chmod(relative_extension, mode)

    print("Copied all compiled dynamic library files into source")


def build() -> None:
    """Construct the extensions and distribution."""  # noqa
    _build_rust_libs()

    # Create C Extensions to feed into cythonize()
    extensions = _build_extensions()
    distribution = _build_distribution(extensions)

    # Build and run the command
    print("Compiling C extension modules...")
    cmd: build_ext = build_ext(distribution)
    if PARALLEL_BUILD:
        cmd.parallel = os.cpu_count()
    cmd.ensure_finalized()
    cmd.run()

    if not SKIP_BUILD_COPY:
        # Copy the build back into the project for development and packaging
        _copy_build_dir_to_project(cmd)


if __name__ == "__main__":
    print("\033[36m")
    print("=====================================================================")
    print("Nautilus Builder")
    print("=====================================================================\033[0m")

    ts_start = datetime.utcnow()

    # Work around a Cython problem in Python 3.8.x on macOS
    # https://github.com/cython/cython/issues/3262
    if platform.system() == "Darwin":
        print("macOS: Setting multiprocessing method to 'fork'.")
        try:
            # noinspection PyUnresolvedReferences
            import multiprocessing

            multiprocessing.set_start_method("fork", force=True)
        except ImportError:  # pragma: no cover
            print("multiprocessing not available")

    # Note: On macOS (and perhaps other platforms), executable files may be
    # universal files containing multiple architectures. To determine the
    # “64-bitness” of the current interpreter, it is more reliable to query the
    # sys.maxsize attribute:
    bits = "64-bit" if sys.maxsize > 2**32 else "32-bit"
    rustc_version = subprocess.check_output(["rustc", "--version"])  # noqa
    print(f"System:  {platform.system()} {bits}")
    print(f"Rust:   {rustc_version.lstrip(b'rustc ').decode()[:-1]}")
    print(f"Python: {platform.python_version()}")
    print(f"Cython: {cython_compiler_version}")
    print(f"NumPy:   {np.__version__}")
    print("")

    print("Starting build...")
    print(f"CARGO_MODE={CARGO_MODE}")
    print(f"DEBUG_MODE={DEBUG_MODE}")
    print(f"ANNOTATION_MODE={ANNOTATION_MODE}")
    print(f"PARALLEL_BUILD={PARALLEL_BUILD}")
    print(f"SKIP_BUILD_COPY={SKIP_BUILD_COPY}")
    print("")

    build()
    print(f"Build time: {datetime.utcnow() - ts_start}")
    print("\033[32m" + "Build completed" + "\033[0m")



from nautilus_trader.execution.reports cimport OrderStatusReport
from nautilus_trader.execution.reports cimport PositionStatusReport
from nautilus_trader.execution.reports cimport TradeReport
from nautilus_trader.live.execution_engine cimport LiveExecutionEngine as NautilusLiveExecutionEngine


# from nacre.execution.reports cimport PositionStatusReport

cdef class LiveExecutionEngine(NautilusLiveExecutionEngine):
    cdef readonly bint reconcile_by_fake_fills

    cdef void _generate_fake_fills(self, PositionStatusReport report) except *



from nautilus_trader.model.data.base import DataType
from nautilus_trader.model.identifiers import ClientId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.trading.strategy import Strategy
from nautilus_trader.trading.strategy import StrategyConfig

from nacre.model.data.tick import MarkTick


class MonConfig(StrategyConfig):
    venues: str  # separate by comma


class Mon(Strategy):
    def __init__(self, config: MonConfig):
        super().__init__(config)
        self.venues = [Venue(v) for v in config.venues.split(",")]

    def on_start(self):
        for venue in self.venues:
            self.subscribe_data(
                client_id=ClientId(venue.value),
                data_type=DataType(MarkTick),
            )
            for id in self.cache.instrument_ids(venue=venue):
                self.subscribe_quote_ticks(id)



import asyncio
from typing import Any, Callable

from nautilus_trader.common.actor cimport Actor
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.common.queue cimport Queue
from nautilus_trader.model.identifiers cimport TraderId

from nacre.config import PubSubConfig


cdef class PubSub(Actor):
    _sentinel = None

    def __init__(self, config: PubSubConfig):
        super().__init__(config)

        self._loop = None
        self.topic_filters = config.topic_filters
        self._queue = Queue(maxsize=10000)  # hardcoded for now

        self._run_queues_task = None
        self.is_running = False

    def get_run_queue_task(self) -> asyncio.Task:
        return self._run_queues_task

    cpdef int qsize(self) except *:
        return self._queue.qsize()

    cpdef void on_start(self) except *:
        self._loop = asyncio.get_running_loop()

        self.is_running = True  # Queues will continue to process

        for topic_filter in self.topic_filters:
            self._msgbus.subscribe(topic=topic_filter, handler=self._handle_data, priority=1)

        # Run queue
        self._run_queues_task = self._loop.create_task(self._run())
        self._log.debug(f"Scheduled {self._run_queues_task}")
        self.connect()

    cpdef void on_stop(self) except *:
        if self.is_running:
            self.is_running = False
            self._enqueue_sentinels()

        self.disconnect()

    def connect(self):
        raise NotImplementedError("method must be implemented in the subclass")  # pragma: no cover

    def disconnect(self):
        raise NotImplementedError("method must be implemented in the subclass")  # pragma: no cover

    cpdef void kill(self) except *:
        self._log.warning("Killing pusub...")
        if self._run_queues_task:
            self._log.debug("Canceling run_queues_task...")
            self._run_queues_task.cancel()
        if self.is_running:
            self.is_running = False  # Avoids sentinel messages for queues
            self.stop()

    cpdef object wrap_message(self, object obj):
        raise NotImplementedError("method must be implemented in the subclass")  # pragma: no cover

    def _handle_data(self, obj):
        message = self.wrap_message(obj)
        try:
            self._queue.put_nowait(message)
        except asyncio.QueueFull:
            self._log.warning(
                f"Blocking on `_queue.put` as message_queue full at "
                f"{self._queue.qsize()} items.",
            )
            self._loop.create_task(self._queue.put(message))  # Blocking until qsize reduces

    async def _run(self):
        self._log.debug(
            f"Message queue processing starting (qsize={self.qsize()})...",
        )
        cdef object message
        cdef str topic
        try:
            while self.is_running:
                message = await self._queue.get()
                if message is None:  # Sentinel message (fast C-level check)
                    continue         # Returns to the top to check `self.is_running`
                else:
                    self.relay(message)

        except asyncio.CancelledError:
            if not self._queue.empty():
                self._log.warning(
                    f"Running canceled with {self.qsize()} message(s) on queue.",
                )
            else:
                self._log.debug(
                    f"Message queue processing stopped (qsize={self.qsize()}).",
                )

        self._log.debug(
            f"Message queue exit (qsize={self.qsize()})...",
        )

    cdef void _enqueue_sentinels(self) except *:
        self._queue.put_nowait(self._sentinel)
        self._log.debug(f"Sentinel message placed on message queue.")

    cpdef void relay(self, object obj) except *:
        """Abstract method (implement in subclass)."""
        raise NotImplementedError("method must be implemented in the subclass")  # pragma: no cover



# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2021 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.network.websocket cimport WebSocketClient as NautilusWebSocketClient


cdef class WebSocketClient(NautilusWebSocketClient):
    cdef readonly bint retrying



import asyncio
import base64
import hashlib
import hmac
from typing import Any, Callable, Dict, Optional
from urllib.parse import urlparse

from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import Logger

from nacre.adapters.zb.common import format_symbol
from nacre.adapters.zb.websocket.client import ZbWebSocketClient


class ZbFuturesWebSocket:
    BASE_URL = "wss://fapi.zb.com/ws/public/v1"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        base_url: Optional[str] = None,
    ):

        usdt_m = ZbFuturesWebsocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=base_url or self.BASE_URL,
        )

        # force path replace
        qc_ws_url = urlparse(base_url or self.BASE_URL)._replace(path="/qc/ws/public/v1").geturl()
        qc_m = ZbFuturesWebsocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=qc_ws_url,
        )

        self.clients = {"usdt": usdt_m, "qc": qc_m}

    @property
    def is_connected(self) -> bool:
        return all(map(lambda c: c.is_connected, self.clients.values()))

    def get_client(self, symbol: str):
        if "/" in symbol:
            quote = symbol.partition("/")[2]
        elif "_" in symbol:
            quote = symbol.partition("_")[2]
        else:
            raise ValueError(f"Symbol {symbol} not supported")
        if quote == "ZUSD":
            quote = "usdt"
        return self.clients[quote.lower()]

    async def connect(self, **kwargs):
        await asyncio.gather(*[client.connect(**kwargs) for client in self.clients.values()])

    async def disconnect(self):
        # very weird bug here, can't await both otherwise it will stuck at return
        await self.clients["usdt"].disconnect()
        # await self.clients["qc"].disconnect()

    async def subscribe_mark_price(self, symbol: Optional[str] = None):
        if symbol:
            client = self.get_client(symbol)
            await client.subscribe_mark_price(symbol)
        else:
            await asyncio.gather(
                *[client.subscribe_mark_price(symbol) for client in self.clients.values()]
            )

    async def subscribe_index_price(self, symbol: Optional[str] = None):
        if symbol:
            client = self.get_client(symbol)
            await client.subscribe_index_price(symbol)
        else:
            await asyncio.gather(
                *[client.subscribe_index_price(symbol) for client in self.clients.values()]
            )

    async def subscribe_mark_bars(self, symbol: str, interval: str, size: int = 1):
        client = self.get_client(symbol)
        await client.subscribe_mark_bars(symbol, interval, size)

    async def subscribe_index_bars(self, symbol: str, interval: str, size: int = 1):
        client = self.get_client(symbol)
        await client.subscribe_index_bars(symbol, interval, size)

    async def subscribe_trades(self, symbol: str, size: int = 50):
        client = self.get_client(symbol)
        await client.subscribe_trades(symbol, size)

    async def subscribe_bars(self, symbol: str, interval: str, size: int = 1):
        client = self.get_client(symbol)
        await client.subscribe_bars(symbol, interval, size)

    async def subscribe_ticker(self, symbol: Optional[str] = None):
        if symbol:
            client = self.get_client(symbol)
            await client.subscribe_ticker(symbol)
        else:
            await asyncio.gather(
                *[client.subscribe_ticker(symbol) for client in self.clients.values()]
            )

    async def subscribe_book_deltas(
        self, symbol: str, depth: int = 50, precision: Optional[float] = None
    ):
        client = self.get_client(symbol)
        await client.subscribe_book_deltas(symbol, depth, precision)

    async def subscribe_book_snapshot(
        self, symbol: str, depth: int = 5, precision: Optional[float] = None
    ):
        client = self.get_client(symbol)
        await client.subscribe_book_snapshot(symbol, depth, precision)


class ZbFuturesWebsocketClient(ZbWebSocketClient):
    """
    Provides access to the `Zb FUTURES` streaming WebSocket API.
    """

    BASE_URL = "wss://fapi.zb.com/ws/public/v1"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        base_url: Optional[str] = None,
    ):
        super().__init__(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=base_url or self.BASE_URL,
        )

    async def _subscribe_channel(self, channel: str, **kwargs):
        kwargs["action"] = "subscribe"

        await super()._subscribe_channel(channel, **kwargs)

    async def _unsubscribe_channel(self, channel: str, **kwargs):
        kwargs["action"] = "unsubscribe"
        await super()._unsubscribe_channel(channel, **kwargs)

    async def subscribe_mark_price(self, symbol: Optional[str] = None):
        chan = "All"
        if symbol:
            chan = format_symbol(symbol)
        await self._subscribe_channel(channel=f"{chan}.mark")

    async def subscribe_index_price(self, symbol: Optional[str] = None):
        chan = "All"
        if symbol:
            chan = format_symbol(symbol)
        await self._subscribe_channel(channel=f"{chan}.index")

    async def subscribe_mark_bars(self, symbol: str, interval: str, size: int = 1):
        channel = f"{format_symbol(symbol)}.mark_{interval}"
        await self._subscribe_channel(channel=channel, size=size)

    async def subscribe_index_bars(self, symbol: str, interval: str, size: int = 1):
        channel = f"{format_symbol(symbol)}.index_{interval}"
        await self._subscribe_channel(channel=channel, size=size)

    async def subscribe_trades(self, symbol: str, size: int = 50):
        """
        Trade Streams.

        The Trade Streams push raw trade information; each trade has a unique buyer and seller.
        Update Speed: Real-time

        """
        await self._subscribe_channel(channel=f"{format_symbol(symbol)}.Trade", size=size)

    async def subscribe_bars(self, symbol: str, interval: str, size: int = 1):
        """
        Subscribe to bar (kline/candlestick) stream.

        The Kline/Candlestick Stream push updates to the current klines/candlestick every second.
        interval:
        1M,5M,15M, 30M, 1H, 6H, 1D, 5D
        Update Speed: 2000ms

        """
        channel = f"{format_symbol(symbol)}.KLine_{interval}"
        await self._subscribe_channel(channel=channel, size=size)

    async def subscribe_ticker(self, symbol: Optional[str] = None):
        """
        Individual symbol or all symbols ticker.

        24hr rolling window ticker statistics for a single symbol.
        These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs.
        Stream Name: <symbol>@ticker or
        Stream Name: !ticker@arr
        Update Speed: 1000ms

        """
        if symbol is None:
            await self._subscribe_channel(channel="All.Ticker")
        else:
            await self._subscribe_channel(channel=f"{format_symbol(symbol)}.Ticker")

    async def subscribe_book_deltas(
        self, symbol: str, depth: int = 50, precision: Optional[float] = None
    ):
        """
        Partial Book Depth Streams.

        Top bids and asks, Valid are min - 5000, default 50
        Update Speed: real time

        """
        channel = f"{format_symbol(symbol)}.Depth"
        if precision:
            channel = f"{format_symbol(symbol)}.Depth@{precision}"
        await self._subscribe_channel(channel=channel, size=depth)

    async def subscribe_book_snapshot(
        self, symbol: str, depth: int = 5, precision: Optional[float] = None
    ):
        """
        Diff book depth stream.

        Top bids and asks, Valid are 5 - 10
        Update Speed: 200ms
        Order book price and quantity depth updates used to locally manage an order book.

        """
        channel = f"{format_symbol(symbol)}.DepthWhole"
        if precision:
            channel = f"{format_symbol(symbol)}.DepthWhole@{precision}"
        await self._subscribe_channel(channel=channel, size=depth)


class ZbFutureUserDataWebSocket:
    BASE_URL = "wss://fapi.zb.com/ws/private/api/v2"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        key: str,
        hashed_secret: str,
        base_url: Optional[str] = None,
    ):
        usdt_m = ZbFuturesUserDataWebSocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            key=key,
            hashed_secret=hashed_secret,
            futures_account_type=1,
            base_url=base_url or self.BASE_URL,
        )

        # force path replace
        qc_ws_url = (
            urlparse(base_url or self.BASE_URL)._replace(path="/qc/ws/private/api/v2").geturl()
        )

        qc_m = ZbFuturesUserDataWebSocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            key=key,
            hashed_secret=hashed_secret,
            futures_account_type=2,
            base_url=qc_ws_url,
        )

        self.clients = {"usdt": usdt_m, "qc": qc_m}

    @property
    def is_connected(self) -> bool:
        return all(map(lambda c: c.is_connected, self.clients.values()))

    async def connect(self, **kwargs):
        await asyncio.gather(*[client.connect(**kwargs) for client in self.clients.values()])

    async def disconnect(self):
        # very weird bug here, can't await both otherwise it will stuck at return
        await self.clients["usdt"].disconnect()
        # await self.clients["qc"].disconnect()
        # for client in self.clients.values():
        #     await client.disconnect()

    # async def logged_in(self):
    #     for client in self.clients.values():
    #         await client.logged_in()

    # def add_post_connect_callbacks(self, *callbacks: Awaitable):
    #     for client in self.clients.values():
    #         client.add_post_connect_callbacks(*callbacks)

    def get_client(self, symbol: str):
        quote = symbol.partition("/")[2]
        if quote == "ZUSD":
            quote = "usdt"
        return self.clients[quote.lower()]

    async def subscribe_funding_update(self, currency: str):
        await asyncio.gather(
            *[client.subscribe_funding_update(currency) for client in self.clients.values()]
        )

    async def subscribe_asset_update(self):
        await asyncio.gather(*[client.subscribe_asset_update() for client in self.clients.values()])

    async def get_asset_snapshot(self, currency: str):
        await asyncio.gather(
            *[client.get_asset_snapshot(currency) for client in self.clients.values()]
        )

    async def subscribe_position_update(self, symbol: Optional[str] = None):
        if symbol is not None:
            client = self.get_client(symbol)
            await client.subscribe_position_update(symbol)
        else:
            await asyncio.gather(
                *[client.subscribe_position_update() for client in self.clients.values()]
            )

    async def subscribe_order_update(self, symbol: Optional[str] = None):
        if symbol is not None:
            client = self.get_client(symbol)
            await client.subscribe_order_update(symbol)
        else:
            await asyncio.gather(
                *[client.subscribe_order_update() for client in self.clients.values()]
            )

    async def new_order(
        self,
        symbol: str,
        side: int,
        amount: float,
        price: Optional[float] = None,
        action: Optional[int] = None,
        client_order_id: Optional[str] = None,
    ):
        client = self.get_client(symbol)
        await client.new_order(symbol, side, amount, price, action, client_order_id)

    async def cancel_order(
        self,
        symbol: str,
        order_id: Optional[str] = None,
        client_order_id: Optional[str] = None,
    ):
        client = self.get_client(symbol)
        await client.cancel_order(symbol, order_id, client_order_id)

    async def cancel_open_orders(self, symbol: str):
        client = self.get_client(symbol)
        await client.cancel_open_orders(symbol)

    async def get_trade_list(self, symbol: str, order_id: str):
        client = self.get_client(symbol)
        await client.get_trade_list(symbol, order_id)


class ZbFuturesUserDataWebSocketClient(ZbWebSocketClient):
    BASE_URL = "wss://fapi.zb.com/ws/private/api/v2"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        key: str,
        hashed_secret: str,
        futures_account_type: int = 1,
        base_url: Optional[str] = None,
    ):
        super().__init__(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=base_url or self.BASE_URL,
        )

        self._key = key
        self._hashed_secret = hashed_secret
        # self.is_logged_in = False
        self.futures_account_type = futures_account_type

    async def _request_channel(self, channel: str, **kwargs):
        kwargs["action"] = "subscribe"

        await super()._request_channel(channel, **kwargs)

    async def _subscribe_channel(self, channel: str, **kwargs):
        kwargs["action"] = "subscribe"

        await super()._subscribe_channel(channel, **kwargs)

    def _get_sign(self, timestamp, http_method, url_path) -> str:
        whole_data = timestamp + http_method + url_path
        m = hmac.new(self._hashed_secret.encode(), whole_data.encode(), hashlib.sha256)
        return str(base64.b64encode(m.digest()), "utf-8")

    async def _login(self):
        """
        Login to the user data stream.

        """
        timestamp = self._clock.utc_now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
        signature = self._get_sign(timestamp, "GET", "login")
        payload = {
            "action": "login",
            "ZB-APIKEY": self._key,
            "ZB-TIMESTAMP": timestamp,
            "ZB-SIGN": signature,
        }

        await self.send_json(payload)

    # async def logged_in(self):
    #     while not self.is_logged_in:
    #         await self._sleep0()
    #     self._log.debug("Websocket logged in")

    async def on_post_connect(self):
        await self._login()
        await super().on_post_connect()
        # self.is_logged_in = True

    async def subscribe_funding_update(self, currency: str):
        await self._subscribe_channel(
            channel="Fund.change", futuresAccountType=self.futures_account_type, currency=currency
        )

    async def subscribe_asset_update(self):
        await self._subscribe_channel(
            channel="Fund.assetChange", futuresAccountType=self.futures_account_type
        )

    async def get_asset_snapshot(self, currency: str):
        await self._request_channel(
            channel="Fund.balance", futuresAccountType=self.futures_account_type, currency=currency
        )

    async def subscribe_position_update(self, symbol: Optional[str] = None):
        payload: Dict[str, Any] = {"futuresAccountType": self.futures_account_type}
        if symbol:
            payload["symbol"] = format_symbol(symbol)

        await self._subscribe_channel(channel="Positions.change", **payload)

    async def subscribe_order_update(self, symbol: Optional[str] = None):
        payload = {}
        if symbol:
            payload["symbol"] = format_symbol(symbol)

        await self._subscribe_channel(channel="Trade.orderChange", **payload)

    async def new_order(
        self,
        symbol: str,
        side: int,
        amount: float,
        price: Optional[float] = None,
        action: Optional[int] = None,
        client_order_id: Optional[str] = None,
    ):
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol), "side": side, "amount": amount}
        if price is not None:
            payload["price"] = price
        if action is not None:
            payload["actionType"] = action
        if client_order_id is not None:
            payload["clientOrderId"] = client_order_id

        await self._request_channel(channel="Trade.order", **payload)

    async def cancel_order(
        self,
        symbol: str,
        order_id: Optional[str] = None,
        client_order_id: Optional[str] = None,
    ):
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if order_id is not None:
            payload["orderId"] = order_id
        elif client_order_id is not None:
            payload["clientOrderId"] = client_order_id

        await self._request_channel(channel="Trade.cancelOrder", **payload)

    async def cancel_open_orders(self, symbol: str):
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        await self._request_channel(channel="Trade.cancelAllOrders", **payload)

    async def get_trade_list(self, symbol: str, order_id: str):
        payload = {}
        payload["symbol"] = format_symbol(symbol)
        payload["orderId"] = order_id
        await self._request_channel(channel="Trade.getTradeList", **payload)



import importlib
import os
import sys
from typing import Dict, List

import fire
import yaml
from nautilus_trader.live.node import TradingNode

from nacre.config.nodes import TradingNodeConfig


def run():
    fire.Fire(run_live_node)


def run_live_node(config: str):
    with open(config, "r") as stream:
        node_cfg = yaml.safe_load(stream)

    # Insert search path
    insert_path(node_cfg.pop("paths", None))

    # Insert actors
    insert_actors(node_cfg)

    # Insert API KEY/SECRET
    insert_credential(node_cfg)

    cfg = TradingNodeConfig.parse_obj(node_cfg)
    node = TradingNode(cfg)

    # Insert Data/Exec factories
    insert_factories(node, node_cfg)

    node.build()
    try:
        node.start()
    finally:
        node.dispose()


def insert_actors(node_cfg):
    actors = build_auxiliary_actors()
    if "actors" in node_cfg:
        node_cfg["actors"].extend(actors)
    else:
        node_cfg["actors"] = actors


# Load api key/secret from ENV variable from API_KEY_{venue}_{id}
# eg: API_KEY_BINANCE_001, API_SECRET_BINANCE_001
def insert_credential(node_cfg):
    for id, item in node_cfg["exec_clients"].items():
        if "-" not in id:
            continue
        venue, _, account_id = id.partition("-")
        api_key = os.getenv(f"API_KEY_{venue.upper()}_{account_id.upper()}")
        api_secret = os.getenv(f"API_SECRET_{venue.upper()}_{account_id.upper()}")

        if "api_key" not in item["config"] and api_key is not None:
            node_cfg["exec_clients"][id]["config"]["api_key"] = api_key
            node_cfg["data_clients"][venue]["config"]["api_key"] = api_key
        if "api_secret" not in item["config"] and api_secret is not None:
            node_cfg["exec_clients"][id]["config"]["api_secret"] = api_secret
            node_cfg["data_clients"][venue]["config"]["api_secret"] = api_secret


def insert_factories(node, node_cfg):
    for venue, item in node_cfg["data_clients"].items():
        node.add_data_client_factory(venue, _import_cls(item["factory_path"]))
    for id, item in node_cfg["exec_clients"].items():
        if "-" in id:
            id = id.split("-")[0]
        node.add_exec_client_factory(id, _import_cls(item["factory_path"]))


def insert_path(paths) -> None:
    if paths and isinstance(paths, list):
        for path in paths:
            sys.path.insert(0, path)


def build_auxiliary_actors() -> List[Dict]:
    actors = []
    exposer_port = os.getenv("METRICS_PORT")
    if exposer_port:
        actors.append(
            {
                "actor_path": "nacre.actors.exposer:Exposer",
                "config_path": "nacre.config.actors:ExposerConfig",
                "config": {
                    "host": "0.0.0.0",  # noqa: S104
                    "port": exposer_port,
                },
            }
        )

    kafka_bootstrap_servers = os.getenv("KAFKA_BOOTSTRAP_SERVERS")
    schema_registry = os.getenv("SCHEMA_REGISTRY")
    if kafka_bootstrap_servers and schema_registry:
        topic_filters = os.getenv("TOPIC_FILTERS")
        if topic_filters is None:
            filters = ["events.snapshot*"]
        else:
            filters = topic_filters.split(",")

        actors.append(
            {
                "actor_path": "nacre.infrastructure.kafka:KafkaPubSub",
                "config_path": "nacre.config.actors:KafkaConfig",
                "config": {
                    "bootstrap_servers": kafka_bootstrap_servers,
                    "schema_registry": schema_registry,
                    "topic_filters": filters,
                },
            }
        )

    return actors


def _import_cls(path: str):
    assert path
    assert ":" in path

    module, cls = path.rsplit(":")
    mod = importlib.import_module(module)
    return getattr(mod, cls)


if __name__ == "__main__":
    run()



from typing import Optional


def format_symbol(symbol: str):
    return symbol.upper().replace("/", "_")


def format_market(market: str):
    return market.lower().replace("/", "_")


def format_websocket_market(market: str):
    return market.lower().replace("/", "").replace("_", "")


def format_websocket_user_market(market: str):
    suffix = "default"
    market = market.lower()
    if market.endswith(suffix):
        return market[: len(market) - len(suffix)]
    return market


def format_endpoint(
    endpoint: str,
    symbol: Optional[str] = None,
    future_account_type: int = 1,
    currency: Optional[str] = None,
) -> str:
    if symbol is not None:
        quote = symbol.partition("/")[2]
        if quote == "ZUSD":
            quote = "usdt"
        return endpoint.format(quote=quote.lower())
    if currency is not None:
        return endpoint.format(quote=currency.lower())

    if future_account_type == 1:
        return endpoint.format(quote="usdt")
    elif future_account_type == 2:
        return endpoint.format(quote="qc")
    else:
        raise ValueError(f"{future_account_type} account type not supported")




from nautilus_trader.execution.reports cimport PositionStatusReport as NautilusPositionStatusReport
from nautilus_trader.model.objects cimport Price


cdef class PositionStatusReport(NautilusPositionStatusReport):
    cdef readonly Price entry_px
    cdef readonly Price last_px



from nautilus_trader.live import node

from nacre.system.kernel import NautilusKernel


node.NautilusKernel = NautilusKernel

from nacre.config.actors import ExposerConfig
from nacre.config.actors import KafkaConfig
from nacre.config.actors import PubSubConfig
from nacre.config.live import LiveExecEngineConfig
from nacre.config.nodes import TradingNodeConfig


__all__ = [
    "ExposerConfig",
    "KafkaConfig",
    "PubSubConfig",
    "LiveExecEngineConfig",
    "TradingNodeConfig",
]



import asyncio
import base64
import hashlib
import hmac
from typing import Any, Dict

import nautilus_trader
import orjson
from aiohttp import ClientConnectorError
from aiohttp import ClientResponse
from aiohttp import ClientResponseError
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import Logger

# from nautilus_trader.network.http import HttpClient
from tenacity import retry
from tenacity.retry import retry_if_exception
from tenacity.stop import stop_after_delay

from nacre.adapters.utils import after_log
from nacre.adapters.zb.http.error import ZbClientError
from nacre.adapters.zb.http.error import ZbOperationError
from nacre.adapters.zb.http.error import ZbServerError
from nacre.adapters.zb.http.error import ZbSpotOperationError
from nacre.network.http import HttpClient


def retry_if_connect_error(exception):
    if isinstance(exception, ZbOperationError):
        if exception.status in [10023, 10026, 10027, 10028, 10029, 12027]:
            return True
        return False
    elif isinstance(exception, ZbSpotOperationError):
        if exception.status in [1002, 3007]:
            return True
        return False

    return isinstance(exception, (ClientConnectorError, ZbClientError))


NAUTILUS_VERSION = nautilus_trader.__version__


class ZbHttpClient(HttpClient):
    """
    Provides a `Zb` asynchronous HTTP client.
    """

    BASE_URL = "https://fapi.zb.com"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        key=None,
        secret=None,
        base_url=None,
        timeout=None,
        show_limit_usage=False,
        proxy=None,
    ):
        super().__init__(
            loop=loop,
            logger=logger,
        )
        self._clock = clock
        self._key = key
        self._hashed_secret = hashlib.sha1(  # noqa: S303, S324
            (secret or "").encode("utf-8")
        ).hexdigest()
        self._base_url = base_url or self.BASE_URL
        self._show_limit_usage = show_limit_usage
        self._proxies = None
        self._headers: Dict[str, str] = {
            "Content-Type": "application/json;charset=utf-8",
            "User-Agent": "nautilus-trader/" + NAUTILUS_VERSION,
            "ZB-APIKEY": key,
            "ZB-LAN": "cn",
        }
        self._timeout = timeout
        self._proxy = proxy

        # TODO(cs): Implement limit usage

    @property
    def api_key(self) -> str:
        return self._key

    @property
    def hashed_secret(self) -> str:
        return self._hashed_secret

    @property
    def headers(self):
        return self._headers

    async def query(self, url_path, payload: Dict[str, str] = None) -> Any:
        return await self.send_request("GET", url_path, payload=payload)

    async def limit_request(
        self,
        http_method: str,
        url_path: str,
        payload: Dict[str, Any] = None,
    ) -> Any:
        """
        Limit request is for those endpoints requiring an API key in the header.
        """
        return await self.send_request(http_method, url_path, payload=payload)

    async def sign_request(
        self,
        http_method: str,
        url_path: str,
        payload: Dict[str, str] = None,
    ) -> Any:
        if payload is None:
            payload = {}
        timestamp = self._clock.utc_now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
        query_string = self._prepare_params(payload)
        signature = self._get_sign(timestamp, http_method, url_path, query_string)
        extra_headers = {}
        extra_headers["ZB-SIGN"] = signature
        extra_headers["ZB-TIMESTAMP"] = timestamp
        return await self.send_request(http_method, url_path, payload, extra_headers)

    @retry(
        reraise=True,
        retry=retry_if_exception(retry_if_connect_error),
        after=after_log(),
        stop=stop_after_delay(15),
    )
    async def send_request(  # noqa: C901
        self,
        http_method: str,
        url_path: str,
        payload: Dict[str, str] = None,
        extra_headers: Dict[str, str] = None,
    ) -> Any:
        # TODO(cs): Uncomment for development
        # print(f"{http_method} {url_path} {payload}")
        self._log.debug(f"{http_method} {url_path} {payload}")
        if payload is None:
            payload = {}

        headers = self._headers
        if extra_headers:
            headers = dict(extra_headers, **headers)
        try:
            kwargs = dict(
                method=http_method,
                url=self._base_url + url_path,
                headers=headers,
                params=self._prepare_params(payload),
            )
            if self._proxy:
                kwargs["proxy"] = self._proxy
            if self._timeout:
                kwargs["timeout"] = self._timeout

            if http_method == "POST":
                kwargs["json"] = payload
                kwargs.pop("params")

            resp: ClientResponse = await self.request(**kwargs)
        except ClientResponseError as ex:
            await self._handle_exception(ex)
            return

        try:
            resp_data = orjson.loads(resp.data)
            self._raise_if_operation_code_error(resp_data)
            return resp_data
        except orjson.JSONDecodeError:
            self._log.error(f"Could not decode data to JSON: {resp.data}.")

    def _prepare_params(self, params: Dict[str, str]) -> str:
        keys = sorted(params)
        return "&".join(
            [k + "=" + str(params[k]) for k in keys if params[k] is not None and params[k] != ""]
        )

    def _get_sign(self, timestamp, http_method, url_path, data) -> str:
        whole_data = timestamp + http_method + url_path + data
        m = hmac.new(self._hashed_secret.encode(), whole_data.encode(), hashlib.sha256)
        return str(base64.b64encode(m.digest()), "utf-8")

    async def _handle_exception(self, error: ClientResponseError) -> None:
        if error.status < 400:
            return
        elif 400 <= error.status < 500:
            raise ZbClientError(
                status=error.status,
                message=error.message,
                headers=error.headers,
            )
        else:
            raise ZbServerError(
                status=error.status,
                message=error.message,
                headers=error.headers,
            )

    def _raise_if_operation_code_error(self, resp: Dict):
        code = resp.get("code")
        if code is None:
            return
        elif code == 10000:
            return
        else:
            raise ZbOperationError(status=code, message=resp.get("desc"))



from typing import Any, Dict, Optional

from nautilus_trader.core.correctness import PyCondition

from nacre.adapters.zb.common import format_market
from nacre.adapters.zb.http.spot import ZbSpotHttpClient


class ZbSpotAccountHttpAPI:
    BASE_ENDPOINT = "/api/"

    def __init__(self, client: ZbSpotHttpClient):
        """
        Initialize a new instance of the ``ZbSpotAccountHttpAPI`` class.

        Parameters
        ----------
        client : ZbSpotHttpClient
            The Binance REST API client.

        """
        PyCondition.not_none(client, "client")

        self.client = client

    async def get_account(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getAccountInfo",
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getAccountInfo",
            payload=payload,
        )

    async def get_user_address(self, currency: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getUserAddress",
            "currency": format_market(currency),
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getUserAddress",
            payload=payload,
        )

    async def get_payin_address(self, currency: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getPayinAddress",
            "currency": format_market(currency),
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getPayinAddress",
            payload=payload,
        )

    async def get_withdraw_address(self, currency: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getWithdrawAddress",
            "currency": format_market(currency),
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getWithdrawAddress",
            payload=payload,
        )

    async def get_withdraw_record(
        self,
        currency: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getWithdrawRecord",
            "currency": format_market(currency),
        }
        if page_index is not None:
            payload["pageIndex"] = str(page_index)
        if page_size is not None:
            payload["pageSize"] = str(page_size)

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getWithdrawRecord",
            payload=payload,
        )

    async def get_charge_record(
        self,
        currency: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getChargeRecord",
            "currency": format_market(currency),
        }
        if page_index is not None:
            payload["pageIndex"] = str(page_index)
        if page_size is not None:
            payload["pageSize"] = str(page_size)

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getChargeRecord",
            payload=payload,
        )

    async def get_fee_info(self, currency: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getFeeInfo",
            "currency": format_market(currency),
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getFeeInfo",
            payload=payload,
        )

    async def withdraw(
        self,
        amount: float,
        currency: str,
        fees: float,
        receive_addr: str,
        safe_pwd: str,
        itransfer: Optional[int] = 1,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "withdraw",
            "amount": amount,
            "currency": format_market(currency),
            "fees": fees,
            "itransfer": itransfer,
            "receiveAddr": receive_addr,
            "safePwd": safe_pwd,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "withdraw",
            payload=payload,
        )

    async def order(
        self,
        amount: float,
        currency: str,
        price: float,
        trade_type: int,
        customer_order_id: Optional[str] = None,
        order_type: Optional[int] = None,
        acct_type: int = 0,
    ) -> Dict[str, Any]:
        """
        Submit a new order.

        Parameters
        ----------
        amount : float
            Amount
        currency : str
            Symbol
        price :  float
            Price
        trade_type : int
            Transaction type, 1 for buy, 0 for sell
        customer_order_id : str, optional
            User defined ID, 4-36 characters, consisting of numbers or letters
        order_type : str, optional
            Delegate type, 1 for PostOnly, 2 for IOC, None for Limit
        acct_type : int, optional
            Mode, 0 for spot, 1 for isolated margin, 2 for cross magin, default spot

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "method": "order",
            "acctType": acct_type,
            "amount": amount,
            "currency": format_market(currency),
            "price": price,
            "tradeType": trade_type,
        }
        if order_type is not None:
            payload["orderType"] = order_type
        if customer_order_id is not None:
            payload["customerOrderId"] = customer_order_id

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "order",
            payload=payload,
        )

    async def order_morev2(
        self,
        market: str,
        trade_params: str,
        trade_type: int,
        acct_type: Optional[int] = 0,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "orderMoreV2",
            "market": market,
            "tradeParams": trade_params,
            "tradeType": trade_type,
            "acctType": acct_type,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "orderMoreV2",
            payload=payload,
        )

    async def cancel_order(
        self, currency: str, id: Optional[int] = None, customer_order_id: Optional[str] = None
    ) -> Dict[str, Any]:
        if (id is not None) == (customer_order_id is not None):
            raise ValueError("pass customer_order_id or id")

        payload: Dict[str, Any] = {
            "method": "cancelOrder",
            "currency": format_market(currency),
        }

        if id is not None:
            payload["id"] = str(id)
        if customer_order_id is not None:
            payload["customerOrderId"] = customer_order_id

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "cancelOrder",
            payload=payload,
        )

    async def get_order(
        self, symbol: str, order_id: Optional[str] = None, client_order_id: Optional[str] = None
    ) -> Dict[str, Any]:
        if (order_id is not None) == (client_order_id is not None):
            raise ValueError("pass client_order_id or order_id")

        payload: Dict[str, str] = {
            "method": "getOrder",
            "currency": format_market(symbol),
        }
        if order_id is not None:
            payload["id"] = order_id
        if client_order_id is not None:
            payload["customerOrderId"] = client_order_id

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getOrder",
            payload=payload,
        )

    async def get_orders(self, trade_type: int, currency: str, page_index: Optional[int] = None):
        payload: Dict[str, Any] = {
            "method": "getOrders",
            "tradeType": str(trade_type),
            "currency": format_market(currency),
        }

        if page_index is not None:
            payload["pageIndex"] = page_index

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getOrders",
            payload=payload,
        )

    async def get_orders_ignore_trade_type(
        self,
        currency: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getOrdersIgnoreTradeType",
            "currency": format_market(currency),
        }
        if page_index is not None:
            payload["pageIndex"] = str(page_index)
        if page_size is not None:
            payload["pageSize"] = str(page_size)

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getOrdersIgnoreTradeType",
            payload=payload,
        )

    async def get_unfinished_orders_ignore_trade_type(
        self,
        currency: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getUnfinishedOrdersIgnoreTradeType",
            "currency": format_market(currency),
        }

        if page_index is not None:
            payload["pageIndex"] = str(page_index)
        if page_size is not None:
            payload["pageSize"] = str(page_size)

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getUnfinishedOrdersIgnoreTradeType",
            payload=payload,
        )

    async def get_finished_and_partial_orders(
        self,
        currency: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getFinishedAndPartialOrders",
            "currency": format_market(currency),
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getFinishedAndPartialOrders",
            payload=payload,
        )

    async def add_sub_user(self, memo: str, password: str, sub_user_name: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "memo": memo,
            "password": password,
            "subUserName": sub_user_name,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "addSubUser",
            payload=payload,
        )

    async def get_sub_user_list(self) -> Dict[str, Any]:
        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getSubUserList",
            payload=None,
        )

    async def do_transfer_funds(
        self, amount: float, currency: str, from_user_name: str, to_user_name: str
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "amount": amount,
            "currency": format_market(currency),
            "fromUserName": from_user_name,
            "toUserName": to_user_name,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "doTransferFunds",
            payload=payload,
        )

    async def get_lever_assets_info(self) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getLeverAssetsInfo",
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getLeverAssetsInfo",
            payload=payload,
        )

    async def get_lever_bills(
        self,
        coin: str,
        data_type: int,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getLeverBills",
            "coin": coin,
            "dataType": data_type,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getLeverBills",
            payload=payload,
        )

    async def transfer_in_lever(self, coin: str, market_name: str, amount: float) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "transferInLever",
            "coin": coin,
            "marketName": market_name,
            "amount": amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "transferInLever",
            payload=payload,
        )

    async def transfer_out_lever(
        self, coin: str, market_name: str, amount: float
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "transferOutLever",
            "coin": coin,
            "marketName": market_name,
            "amount": amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "transferOutLever",
            payload=payload,
        )

    async def loan(
        self,
        coin: str,
        amount: float,
        interest_rate_of_day: float,
        repayment_day: int,
        is_loop: bool,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "load",
            "coin": coin,
            "amount": amount,
            "interestRateOfDay": interest_rate_of_day,
            "repaymentDay": repayment_day,
            "isLoop": is_loop,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "loan",
            payload=payload,
        )

    async def cancel_loan(self, loan_id: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "cancelLoan",
            "loanId": loan_id,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "cancelLoan",
            payload=payload,
        )

    async def get_loans(
        self,
        coin: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getLoans",
            "coin": coin,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getLoans",
            payload=payload,
        )

    async def get_loan_records(
        self,
        market_name: str,
        status: int,
        loan_id: Optional[str] = "",
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getLoanRecords",
            "loanId": loan_id,
            "marketName": market_name,
            "status": status,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getLoanRecords",
            payload=payload,
        )

    async def borrow(
        self,
        market_name: str,
        coin: str,
        amount: float,
        interest_rate_of_day: float,
        repayment_day: int,
        is_loop: int,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "borrow",
            "marketName": market_name,
            "coin": coin,
            "amount": amount,
            "interestRateOfDay": str(interest_rate_of_day),
            "repaymentDay": str(repayment_day),
            "isLoop": is_loop,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "borrow",
            payload=payload,
        )

    async def auto_borrow(
        self,
        market_name: str,
        coin: str,
        amount: float,
        interest_rate_of_day: float,
        repayment_day: int,
        safe_pwd: str,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "autoBorrow",
            "marketName": market_name,
            "coin": coin,
            "amount": amount,
            "interestRateOfDay": interest_rate_of_day,
            "repaymentDay": repayment_day,
            "safePwd": safe_pwd,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "autoBorrow",
            payload=payload,
        )

    async def repay(
        self, loan_record_id: str, repay_amount: float, repay_type: int
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "repay",
            "loanRecordId": loan_record_id,
            "repayAmount": repay_amount,
            "repayType": repay_type,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "repay",
            payload=payload,
        )

    async def do_all_repay(self, market_name: str, coin: str, repay_amount: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "doAllRepay",
            "marketName": market_name,
            "coin": coin,
            "repayAmount": repay_amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "doAllRepay",
            payload=payload,
        )

    async def get_repayments(
        self,
        loan_record_id: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getRepayments",
            "loanRecordId": loan_record_id,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getRepayments",
            payload=payload,
        )

    async def get_finance_records(
        self,
        coin: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getFinanceRecords",
            "coin": coin,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getFinanceRecords",
            payload=payload,
        )

    async def change_invest_mark(self, invest_mark: int, loan_record_id: str) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "changeInvestMark",
            "investMark": invest_mark,
            "loanRecordId": loan_record_id,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "changeInvestMark",
            payload=payload,
        )

    async def change_loop(self, is_loop: int, loan_id: str) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "changeLoop",
            "isLoop": is_loop,
            "loanId": loan_id,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "changeLoop",
            payload=payload,
        )

    async def get_cross_assets(self) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "method": "getCrossAssets",
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getCrossAssets",
            payload=payload,
        )

    async def get_cross_bills(
        self,
        coin: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getCrossBills",
            "coin": coin,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getCrossBills",
            payload=payload,
        )

    async def transfer_in_cross(self, coin: str, amount: float) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "transferInCross",
            "coin": coin,
            "amount": amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "transferInCross",
            payload=payload,
        )

    async def transfer_out_cross(self, coin: str, amount: float) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "transferOutCross",
            "coin": coin,
            "amount": amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "transferOutCross",
            payload=payload,
        )

    async def do_cross_loan(self, coin: str, amount: float, safe_pwd: str) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "doCrossLoan",
            "coin": coin,
            "amount": amount,
            "safePwd": safe_pwd,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "doCrossLoan",
            payload=payload,
        )

    async def do_cross_repay(self, coin: str, repay_amount: float) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "doCrossRepay",
            "coin": coin,
            "repayAmount": repay_amount,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "doCrossRepay",
            payload=payload,
        )

    async def get_cross_repay_records(
        self,
        coin: str,
        page_index: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "method": "getCrossRepayRecords",
            "coin": coin,
        }

        if page_index is not None:
            payload["pageIndex"] = page_index
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=self.BASE_ENDPOINT + "getCrossRepayRecords",
            payload=payload,
        )



import asyncio
import socket
import urllib.parse
from ssl import SSLContext
from typing import Dict, List, Optional, Union

import aiohttp
import cython
from aiohttp import ClientResponse
from aiohttp import ClientResponseError
from aiohttp import ClientSession
from aiohttp import Fingerprint

from nacre.metrics.metrics import HTTP_ERROR_COUNTER
from nacre.metrics.metrics import REQ_TIME

from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.network.http cimport HttpClient as NautilusHttpClient


cdef int ONE_DAY = 86_400


cdef class HttpClient(NautilusHttpClient):
    def raise_for_status(self, resp: ClientResponse):
        if not resp.ok:
            # reason should always be not None for a started response
            assert resp.reason is not None
            resp.release()

            message = resp.reason
            if resp.data is not None:
                message = resp.data.decode(resp.get_encoding())
            raise ClientResponseError(
                resp.request_info,
                resp.history,
                status=resp.status,
                message=message,
                headers=resp.headers,
            )

    async def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]]=None,
        json: Optional[Dict[str, str]]=None,
        **kwargs,
    ) -> ClientResponse:
        session: ClientSession = self._get_session()
        if session.closed:
            self._log.warning("Session closed: reconnecting.")
            await self.connect()


        with REQ_TIME.labels(method=method, endpoint=url).time():
            with HTTP_ERROR_COUNTER.labels(
                method=method, endpoint=url
            ).count_exceptions():
                async with session.request(
                    method=method,
                    url=url,
                    headers=headers,
                    json=json,
                    **kwargs
                ) as resp:
                    # Get data first
                    resp.data = await resp.read()
                    self.raise_for_status(resp)
                    return resp



import asyncio
from typing import Any, Dict, List, Optional, Union

import orjson
import pandas as pd
from nautilus_trader.cache.cache import Cache
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import Logger
from nautilus_trader.core.correctness import PyCondition
from nautilus_trader.core.uuid import UUID4
from nautilus_trader.live.data_client import LiveMarketDataClient
from nautilus_trader.model.data.bar import BarType
from nautilus_trader.model.data.base import DataType
from nautilus_trader.model.data.tick import QuoteTick
from nautilus_trader.model.data.tick import TradeTick
from nautilus_trader.model.enums import BarAggregation
from nautilus_trader.model.enums import BarAggregationParser
from nautilus_trader.model.enums import BookType
from nautilus_trader.model.enums import PriceType
from nautilus_trader.model.identifiers import ClientId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity
from nautilus_trader.model.orderbook.data import OrderBookDeltas
from nautilus_trader.model.orderbook.data import OrderBookSnapshot
from nautilus_trader.msgbus.bus import MessageBus

from nacre.adapters.zb.data_types import ZbBar
from nacre.adapters.zb.data_types import ZbTicker
from nacre.adapters.zb.http.api.future_market import ZbFutureMarketHttpAPI
from nacre.adapters.zb.http.api.spot_market import ZbSpotMarketHttpAPI
from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.http.error import ZbError
from nacre.adapters.zb.http.spot import ZbSpotHttpClient
from nacre.adapters.zb.parsing import parse_bar
from nacre.adapters.zb.parsing import parse_bar_ws
from nacre.adapters.zb.parsing import parse_book_snapshot_ws
from nacre.adapters.zb.parsing import parse_diff_depth_stream_ws
from nacre.adapters.zb.parsing import parse_mark_price
from nacre.adapters.zb.parsing import parse_spot_bar
from nacre.adapters.zb.parsing import parse_spot_book_snapshot_ws
from nacre.adapters.zb.parsing import parse_spot_ticker_ws
from nacre.adapters.zb.parsing import parse_spot_trade_tick_ws
from nacre.adapters.zb.parsing import parse_ticker_ws
from nacre.adapters.zb.parsing import parse_trade_tick
from nacre.adapters.zb.parsing import parse_trade_tick_ws
from nacre.adapters.zb.providers import ZbInstrumentProvider
from nacre.adapters.zb.websocket.futures import ZbFuturesWebsocketClient
from nacre.adapters.zb.websocket.spot import ZbSpotWebSocket
from nacre.model.data.tick import MarkTick


class ZbDataClient(LiveMarketDataClient):
    def connect(self):
        """
        Connect the client to Zb.
        """
        self._log.info("Connecting...")
        self._loop.create_task(self._connect())

    def disconnect(self):
        """
        Disconnect the client from Zb.
        """
        self._log.info("Disconnecting...")
        self._loop.create_task(self._disconnect())

    async def _connect(self):
        # Connect HTTP client
        if not self._client.connected:
            await self._client.connect()
        try:
            await self._instrument_provider.initialize()
        except ZbError as ex:
            self._log.exception("Error on connecting", ex)
            return

        self._send_all_instruments_to_data_engine()
        self._update_instruments_task = self._loop.create_task(self._update_instruments())

        # Connect WebSocket clients
        self._loop.create_task(self._connect_websockets())

        self._set_connected(True)
        self._log.info("Connected.")

    async def _connect_websockets(self):
        self._log.info("Awaiting subscriptions...")
        await self._ws_api.connect()

    async def _update_instruments(self):
        while True:
            self._log.debug(
                f"Scheduled `update_instruments` to run in "
                f"{self._update_instruments_interval}s."
            )
            await asyncio.sleep(self._update_instruments_interval)
            await self._instrument_provider.load_all_async()
            self._send_all_instruments_to_data_engine()

    async def _disconnect(self):
        # Cancel tasks
        if self._update_instruments_task:
            self._log.debug("Canceling `update_instruments` task...")
            self._update_instruments_task.cancel()

        # Disconnect HTTP client
        if self._client.connected:
            await self._client.disconnect()

        # Disconnect WebSocket clients
        if self._ws_api.is_connected:
            await self._ws_api.disconnect()

        self._set_connected(False)
        self._log.info("Disconnected.")

    # -- SUBSCRIPTIONS -----------------------------------------------------------------------------

    def subscribe_instruments(self):
        """
        Subscribe to instrument data for the venue.

        """
        for instrument_id in list(self._instrument_provider.get_all().keys()):
            self._add_subscription_instrument(instrument_id)

    def subscribe_instrument(self, instrument_id: InstrumentId):
        """
        Subscribe to instrument data for the given instrument ID.

        Parameters
        ----------
        instrument_id : InstrumentId
            The instrument ID to subscribe to.

        """
        self._add_subscription_instrument(instrument_id)

    def subscribe_ticker(self, instrument_id: InstrumentId):
        self._loop.create_task(self._ws_api.subscribe_ticker(instrument_id.symbol.value))
        self._add_subscription_ticker(instrument_id)

    def subscribe_trade_ticks(self, instrument_id: InstrumentId):
        self._loop.create_task(self._ws_api.subscribe_trades(instrument_id.symbol.value))
        self._add_subscription_trade_ticks(instrument_id)

    def subscribe_instrument_status_updates(self, instrument_id: InstrumentId):
        self._log.warning(
            "Cannot subscribe to instrument status updates: "
            "Not currently supported for the Zb integration.",
        )

    def subscribe_instrument_close_prices(self, instrument_id: InstrumentId):
        self._log.warning(
            "Cannot subscribe to instrument status updates: "
            "Not currently supported for the Zb integration.",
        )

    def unsubscribe_instruments(self):
        """
        Unsubscribe from instrument data for the venue.

        """
        for instrument_id in list(self._instrument_provider.get_all().keys()):
            self._remove_subscription_instrument(instrument_id)

    def unsubscribe_instrument(self, instrument_id: InstrumentId):
        """
        Unsubscribe from instrument data for the given instrument ID.

        Parameters
        ----------
        instrument_id : InstrumentId
            The instrument ID to unsubscribe from.

        """
        self._remove_subscription_instrument(instrument_id)

    def unsubscribe_order_book_deltas(self, instrument_id: InstrumentId):
        self._remove_subscription_order_book_deltas(instrument_id)

    def unsubscribe_order_book_snapshots(self, instrument_id: InstrumentId):
        self._remove_subscription_order_book_snapshots(instrument_id)

    def unsubscribe_ticker(self, instrument_id: InstrumentId):
        self._remove_subscription_ticker(instrument_id)

    def unsubscribe_quote_ticks(self, instrument_id: InstrumentId):
        self._remove_subscription_quote_ticks(instrument_id)

    def unsubscribe_trade_ticks(self, instrument_id: InstrumentId):
        self._remove_subscription_trade_ticks(instrument_id)

    def unsubscribe_bars(self, bar_type: BarType):
        self._remove_subscription_bars(bar_type)

    def unsubscribe_instrument_status_updates(self, instrument_id: InstrumentId):
        self._remove_subscription_instrument_status_updates(instrument_id)

    def unsubscribe_instrument_close_prices(self, instrument_id: InstrumentId):
        self._remove_subscription_instrument_close_prices(instrument_id)

    # -- REQUESTS ----------------------------------------------------------------------------------

    def request_quote_ticks(
        self,
        instrument_id: InstrumentId,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        self._log.error(
            "Cannot request historical quote ticks: not published by Zb.",
        )

    def request_bars(
        self,
        bar_type: BarType,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        if bar_type.is_internally_aggregated():
            self._log.error(
                f"Cannot request {bar_type}: "
                f"only historical bars with EXTERNAL aggregation available from Zb.",
            )
            return

        if not bar_type.spec.is_time_aggregated():
            self._log.error(
                f"Cannot request {bar_type}: only time bars are aggregated by Zb.",
            )
            return

        if bar_type.spec.aggregation == BarAggregation.SECOND:
            self._log.error(
                f"Cannot request {bar_type}: second bars are not aggregated by Zb.",
            )
            return

        if bar_type.spec.price_type != PriceType.LAST:
            self._log.error(
                f"Cannot request {bar_type}: "
                f"only historical bars for LAST price type available from Zb.",
            )
            return

        self._loop.create_task(
            self._request_bars(
                bar_type=bar_type,
                from_datetime=from_datetime,
                to_datetime=to_datetime,
                limit=limit,
                correlation_id=correlation_id,
            )
        )

    def _send_all_instruments_to_data_engine(self):
        for instrument in self._instrument_provider.get_all().values():
            self._handle_data(instrument)

        for currency in self._instrument_provider.currencies().values():
            self._cache.add_currency(currency)

    def _handle_pong(self):
        # reset ping timer
        pass


class ZbSpotDataClient(ZbDataClient):
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        client: ZbSpotHttpClient,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: Logger,
        venue: Venue,
        instrument_provider: ZbInstrumentProvider,
        base_url_ws: Optional[str] = None,
    ):
        # HTTP API
        self._market_api = ZbSpotMarketHttpAPI(client=client)
        # websocket api
        self._ws_api = ZbSpotWebSocket(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=self._handle_ws_message,
            base_url=base_url_ws,
        )

        super(ZbDataClient, self).__init__(
            loop=loop,
            client_id=ClientId(venue.value),
            venue=venue,
            instrument_provider=instrument_provider,
            msgbus=msgbus,
            cache=cache,
            clock=clock,
            logger=logger,
            config={"name": f"ZbDataClient-{venue.value}"},
        )

        self._client = client

        self._quote_tick_subscribed = False
        self._last_quote_tick: Dict[InstrumentId, QuoteTick] = {}
        self._update_instrument_interval: int = 60 * 60  # Once per hour (hardcode)
        self._update_instruments_task: Optional[asyncio.Task] = None

    def subscribe_order_book_deltas(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
        kwargs: dict = None,
    ):
        self._loop.create_task(
            self._subscribe_order_book_snapshot(
                instrument_id=instrument_id,
                book_type=book_type,
                depth=depth,
            )
        )

        self._add_subscription_order_book_deltas(instrument_id)

    def subscribe_order_book_snapshots(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
        kwargs: dict = None,
    ):
        self._loop.create_task(
            self._subscribe_order_book_snapshot(
                instrument_id=instrument_id,
                book_type=book_type,
                depth=depth,
            )
        )

        self._add_subscription_order_book_snapshots(instrument_id)

    async def _subscribe_order_book_snapshot(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
    ):
        if book_type == BookType.L3_MBO:
            self._log.error(
                "Cannot subscribe to order book deltas: "
                "L3_MBO data is not published by Zb. "
                "Valid book types are L1_TBBO, L2_MBP.",
            )
            return

        if depth is not None and depth != 0:
            self._log.warning(
                "Cannot subscribe to depth order book snapshots: "
                f"invalid depth, was {depth}. "
                "Zb does not support custom depth.",
            )

        await self._ws_api.subscribe_depth(
            symbol=instrument_id.symbol.value,
        )

    def subscribe_quote_ticks(self, instrument_id: InstrumentId):
        self._loop.create_task(
            self._ws_api.subscribe_depth(
                symbol=instrument_id.symbol.value,
            )
        )
        self._quote_tick_subscribed = True
        self._add_subscription_quote_ticks(instrument_id)

    def subscribe_bars(self, bar_type: BarType):
        self._log.error(
            "Cannot subscribe to bar updates: "
            "Not currently supported for the Zb spot integration.",
        )

    def request_trade_ticks(
        self,
        instrument_id: InstrumentId,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        if limit == 0 or limit > 50:
            limit = 50

        if from_datetime is not None or to_datetime is not None:
            self._log.warning(
                "Trade ticks have been requested with a from/to time range, "
                f"however the request will be for the most recent {limit}."
            )

        self._loop.create_task(self._request_trade_ticks(instrument_id, limit, correlation_id))

    async def _request_trade_ticks(
        self,
        instrument_id: InstrumentId,
        limit: int,
        correlation_id: UUID4,
    ):
        response: Dict[str, Any] = await self._market_api.trades(instrument_id.symbol.value, limit)

        ticks: List[TradeTick] = [
            parse_trade_tick(
                msg=t,
                instrument_id=instrument_id,
                ts_init=self._clock.timestamp_ns(),
            )
            for t in response["data"]
        ]

        self._handle_trade_ticks(instrument_id, ticks, correlation_id)

    async def _request_bars(
        self,
        bar_type: BarType,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        if limit == 0 or limit > 1000:
            limit = 1000

        if bar_type.spec.aggregation == BarAggregation.MINUTE:
            resolution = "min"
        elif bar_type.spec.aggregation == BarAggregation.HOUR:
            resolution = "hour"
        elif bar_type.spec.aggregation == BarAggregation.DAY:
            resolution = "day"
        else:  # pragma: no cover (design-time error)
            raise RuntimeError(
                f"invalid aggregation period, "
                f"was {BarAggregationParser.from_str(bar_type.spec.aggregation)}",
            )

        if to_datetime is not None:
            self._log.warning(
                "Trade ticks have been requested with to_datetime option, "
                f"however the request will be for the most recent {limit} or after from_datetime."
            )

        start_time_ms = from_datetime.to_datetime64() * 1000 if from_datetime is not None else None

        data: Dict[str, Any] = await self._market_api.kline(
            market=bar_type.instrument_id.symbol.value,
            type=f"{bar_type.spec.step}{resolution}",
            since=start_time_ms,
            size=limit,
        )

        bars: List[ZbBar] = [
            parse_spot_bar(bar_type, values=b, ts_init=self._clock.timestamp_ns())
            for b in data["data"]
        ]
        partial: ZbBar = bars.pop()

        self._handle_bars(bar_type, bars, partial, correlation_id)

    def _handle_ws_message(self, raw: bytes):
        try:
            msg: Dict[str, Any] = orjson.loads(raw)
        except orjson.JSONDecodeError as ex:
            if raw.decode() == "pong":
                self._handle_pong()
                return
            else:
                raise ex

        # data = msg.get("data")
        chan: Optional[str] = msg.get("channel")

        if not chan:
            self._log.error(f"Unknown response: {msg}")
            return

        symbol, _, chan_type = chan.partition("_")

        if chan_type == "ticker":
            self._handle_24hr_ticker(msg["ticker"], symbol)
        elif chan_type == "trades":
            self._handle_trade(msg["data"], symbol)
        elif chan_type == "depth":
            self._handle_book_snapshot(msg, symbol)
        else:
            self._log.error(f"Unrecognized websocket message type, was {chan}")
            return

    def _handle_24hr_ticker(self, data: Dict[str, str], symbol: str):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_market_id(symbol)
        ticker: ZbTicker = parse_spot_ticker_ws(
            instrument_id=instrument_id,
            msg=data,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(ticker)

    def _handle_trade(self, data: List[Dict[str, str]], symbol: str):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_market_id(symbol)
        for d in data:
            trade_tick: TradeTick = parse_spot_trade_tick_ws(
                instrument_id=instrument_id,
                msg=d,
                ts_init=self._clock.timestamp_ns(),
            )
            self._handle_data(trade_tick)

    def _handle_book_snapshot(
        self,
        data: Dict,
        symbol: str,
    ):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_market_id(symbol)
        book_snapshot: OrderBookSnapshot = parse_spot_book_snapshot_ws(
            instrument_id=instrument_id,
            msg=data,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(book_snapshot)

        if self._quote_tick_subscribed:
            self._handle_quote_tick_from_snapshot(data, instrument_id)

    def _handle_quote_tick_from_snapshot(self, data: Dict[str, Any], instrument_id: InstrumentId):
        bid = data["bids"][0]
        ask = data["asks"][-1]
        quote_tick = QuoteTick(
            instrument_id=instrument_id,
            bid=Price.from_str(str(bid[0])),
            ask=Price.from_str(str(ask[0])),
            bid_size=Quantity.from_str(str(bid[1])),
            ask_size=Quantity.from_str(str(ask[1])),
            ts_event=self._clock.timestamp_ns(),
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(quote_tick)
        self._last_quote_tick[instrument_id] = quote_tick


class ZbFutureDataClient(ZbDataClient):
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        client: ZbHttpClient,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: Logger,
        venue: Venue,
        instrument_provider: ZbInstrumentProvider,
        base_url_ws: Optional[str] = None,
    ):
        # HTTP API
        self._market_api = ZbFutureMarketHttpAPI(client=client)
        # websocket api
        self._ws_api = ZbFuturesWebsocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=self._handle_ws_message,
            base_url=base_url_ws,
        )

        super(ZbDataClient, self).__init__(
            loop=loop,
            client_id=ClientId(venue.value),
            venue=venue,
            instrument_provider=instrument_provider,
            msgbus=msgbus,
            cache=cache,
            clock=clock,
            logger=logger,
            config={"name": f"ZbDataClient-{venue.value}"},
        )

        self._client = client

        self._update_instrument_interval: int = 60 * 60  # Once per hour (hardcode)
        self._update_instruments_task: Optional[asyncio.Task] = None

        self._kline_buffer: Dict[InstrumentId, List[Any]] = {}
        self._quote_tick_subscribed = False
        self._last_quote_tick: Dict[InstrumentId, QuoteTick] = {}

    def subscribe(self, data_type: DataType) -> None:
        if data_type.type == MarkTick:
            instrument_id = data_type.metadata.get("instrument_id")
            if instrument_id:
                self._loop.create_task(self._subscribe_mark_price())
                self._add_subscription(
                    DataType(MarkTick, metadata={"instrument_id": instrument_id})
                )
            else:
                self._loop.create_task(self._subscribe_mark_price(instrument_id))
                self._add_subscription(DataType(MarkTick))

    async def _subscribe_mark_price(self, instrument_id: Optional[InstrumentId] = None):
        if instrument_id:
            await self._ws_api.subscribe_mark_price(symbol=instrument_id.symbol.value)
        else:
            await self._ws_api.subscribe_mark_price()

    def subscribe_quote_ticks(self, instrument_id: InstrumentId):
        self._loop.create_task(
            self._ws_api.subscribe_book_deltas(
                symbol=instrument_id.symbol.value,
                depth=1,
            )
        )
        self._quote_tick_subscribed = True
        self._add_subscription_quote_ticks(instrument_id)

    def subscribe_bars(self, bar_type: BarType):
        PyCondition.true(bar_type.is_externally_aggregated(), "aggregation_source is not EXTERNAL")

        if not bar_type.spec.is_time_aggregated():
            self._log.error(
                f"Cannot subscribe to {bar_type}: only time bars are aggregated by Zb.",
            )
            return

        if bar_type.spec.aggregation == BarAggregation.SECOND:
            self._log.error(
                f"Cannot subscribe to {bar_type}: second bars are not aggregated by Zb.",
            )
            return

        if bar_type.spec.aggregation == BarAggregation.MINUTE:
            resolution = "M"
        elif bar_type.spec.aggregation == BarAggregation.HOUR:
            resolution = "H"
        elif bar_type.spec.aggregation == BarAggregation.DAY:
            resolution = "D"
        else:  # pragma: no cover (design-time error)
            raise RuntimeError(
                f"invalid aggregation period, "
                f"was {BarAggregationParser.from_str(bar_type.spec.aggregation)}",
            )

        if bar_type.instrument_id not in self._kline_buffer:
            self._kline_buffer[bar_type.instrument_id] = []

        self._loop.create_task(
            self._ws_api.subscribe_bars(
                symbol=bar_type.instrument_id.symbol.value,
                interval=f"{bar_type.spec.step}{resolution}",
            )
        )
        self._add_subscription_bars(bar_type)

    def subscribe_order_book_deltas(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
        kwargs: dict = None,
    ):
        self._loop.create_task(
            self._subscribe_order_book_deltas(
                instrument_id=instrument_id,
                book_type=book_type,
                depth=depth,
            )
        )

        self._add_subscription_order_book_deltas(instrument_id)

    def subscribe_order_book_snapshots(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
        kwargs: dict = None,
    ):
        self._loop.create_task(
            self._subscribe_order_book_snapshot(
                instrument_id=instrument_id,
                book_type=book_type,
                depth=depth,
            )
        )

        self._add_subscription_order_book_snapshots(instrument_id)

    async def _subscribe_order_book_snapshot(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
    ):
        if book_type == BookType.L3_MBO:
            self._log.error(
                "Cannot subscribe to order book deltas: "
                "L3_MBO data is not published by Zb. "
                "Valid book types are L1_TBBO, L2_MBP.",
            )
            return

        if depth is None or depth == 0:
            depth = 5

        if depth > 10:
            self._log.error(
                "Cannot subscribe to order book snapshots: "
                f"invalid depth, was {depth}. "
                "Valid depths are 1 to 10.",
            )
            return

        await self._ws_api.subscribe_book_snapshot(
            symbol=instrument_id.symbol.value,
            depth=depth,
        )

    async def _subscribe_order_book_deltas(
        self,
        instrument_id: InstrumentId,
        book_type: BookType,
        depth: Optional[int] = None,
    ):
        if book_type == BookType.L3_MBO:
            self._log.error(
                "Cannot subscribe to order book deltas: "
                "L3_MBO data is not published by Zb. "
                "Valid book types are L1_TBBO, L2_MBP.",
            )
            return

        if depth is None or depth == 0:
            depth = 50

        if depth >= 1000:
            self._log.error(
                "Cannot subscribe to order book deltas: "
                f"invalid depth, was {depth}. "
                "Valid depths are 1 to 1000.",
            )
            return

        await self._ws_api.subscribe_book_deltas(
            symbol=instrument_id.symbol.value,
            depth=depth,
        )

    def request_trade_ticks(
        self,
        instrument_id: InstrumentId,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        if limit == 0 or limit > 1000:
            limit = 1000

        if from_datetime is not None or to_datetime is not None:
            self._log.warning(
                "Trade ticks have been requested with a from/to time range, "
                f"however the request will be for the most recent {limit}."
            )

        self._loop.create_task(self._request_trade_ticks(instrument_id, limit, correlation_id))

    async def _request_trade_ticks(
        self,
        instrument_id: InstrumentId,
        limit: int,
        correlation_id: UUID4,
    ):
        response: Dict[str, Any] = await self._market_api.trades(instrument_id.symbol.value, limit)

        ticks: List[TradeTick] = [
            parse_trade_tick(
                msg=t,
                instrument_id=instrument_id,
                ts_init=self._clock.timestamp_ns(),
            )
            for t in response["data"]
        ]

        self._handle_trade_ticks(instrument_id, ticks, correlation_id)

    async def _request_bars(
        self,
        bar_type: BarType,
        from_datetime: pd.Timestamp,
        to_datetime: pd.Timestamp,
        limit: int,
        correlation_id: UUID4,
    ):
        if limit == 0 or limit > 1440:
            limit = 1440

        if bar_type.spec.aggregation == BarAggregation.MINUTE:
            resolution = "M"
        elif bar_type.spec.aggregation == BarAggregation.HOUR:
            resolution = "H"
        elif bar_type.spec.aggregation == BarAggregation.DAY:
            resolution = "D"
        else:  # pragma: no cover (design-time error)
            raise RuntimeError(
                f"invalid aggregation period, "
                f"was {BarAggregationParser.from_str(bar_type.spec.aggregation)}",
            )

        if from_datetime is not None or to_datetime is not None:
            self._log.warning(
                "Trade ticks have been requested with a from/to time range, "
                f"however the request will be for the most recent {limit}."
            )

        data: Dict[str, Any] = await self._market_api.klines(
            symbol=bar_type.instrument_id.symbol.value,
            period=f"{bar_type.spec.step}{resolution}",
            size=limit,
        )

        bars: List[ZbBar] = [
            parse_bar(bar_type, values=b, ts_init=self._clock.timestamp_ns()) for b in data["data"]
        ]
        partial: ZbBar = bars.pop()

        self._handle_bars(bar_type, bars, partial, correlation_id)

    def _handle_ws_message(self, raw: bytes):  # noqa: C901
        msg: Dict[str, Any] = orjson.loads(raw)
        data = msg.get("data")
        chan: Optional[str] = msg.get("channel")
        err: Optional[str] = msg.get("errorCode")

        if err:
            self._log.error(f"Subscribe failed for channel: {chan}, error: {err}")
            return

        if not chan:
            if msg.get("action", "") == "pong":
                self._handle_pong()
            else:
                self._log.error(f"Unknown response: {msg}")
            return

        msg_type: Optional[str] = msg.get("type")

        symbol, _, chan_type = chan.partition(".")

        if chan_type.startswith("DepthWhole"):
            self._handle_book_snapshot(data, symbol)
        elif chan_type.startswith("Depth"):
            if msg_type == "Whole":
                self._handle_book_snapshot(data, symbol)
            else:
                self._handle_depth_update(data, symbol)
        elif chan_type.startswith("KLine"):
            self._handle_kline(data, symbol, chan_type, msg_type)
        elif chan_type == "Trade":
            self._handle_trade(data, symbol, msg_type)
        elif chan_type == "Ticker":
            self._handle_24hr_ticker(data, symbol)
        elif chan_type == "mark":
            self._handle_mark_price(data, symbol)
        elif chan_type == "index":
            self._handle_index_price(data, symbol)
        elif chan_type == "FundingRate":
            self._handle_funding_rate(data, symbol)
        elif chan_type.startswith("mark_"):
            self._handle_mark_kline(data, symbol, chan_type, msg_type)
        elif chan_type.startswith("index_"):
            self._handle_index_kline(data, symbol, chan_type, msg_type)
        else:
            self._log.error(f"Unrecognized websocket message type, was {chan}")
            return

    def _handle_book_snapshot(
        self,
        data: Dict[str, Any],
        symbol: str,
    ):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        book_snapshot: OrderBookSnapshot = parse_book_snapshot_ws(
            instrument_id=instrument_id,
            msg=data,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(book_snapshot)

        if self._quote_tick_subscribed:
            self._handle_quote_tick_from_snapshot(data, instrument_id)

    def _handle_quote_tick_from_snapshot(self, data: Dict[str, Any], instrument_id: InstrumentId):
        bid = data["bids"][0]
        ask = data["asks"][0]
        quote_tick = QuoteTick(
            instrument_id=instrument_id,
            bid=Price.from_str(str(bid[0])),
            ask=Price.from_str(str(ask[0])),
            bid_size=Quantity.from_str(str(bid[1])),
            ask_size=Quantity.from_str(str(ask[1])),
            ts_event=self._clock.timestamp_ns(),
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(quote_tick)
        self._last_quote_tick[instrument_id] = quote_tick

    def _handle_depth_update(
        self,
        data: Dict[str, Any],
        symbol: str,
    ):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        book_deltas: OrderBookDeltas = parse_diff_depth_stream_ws(
            instrument_id=instrument_id,
            msg=data,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(book_deltas)
        if self._quote_tick_subscribed:
            self._handle_quote_tick_from_deltas(data, instrument_id)

    def _handle_quote_tick_from_deltas(self, data: Dict[str, Any], instrument_id: InstrumentId):
        bid, bid_size = next((x for x in data.get("bids", []) if float(x[1]) != 0), [0, 0])
        ask, ask_size = next((x for x in data.get("asks", []) if float(x[1]) != 0), [0, 0])

        bid = self._last_quote_tick[instrument_id].bid if bid == 0 else Price.from_str(str(bid))
        bid_size = (
            self._last_quote_tick[instrument_id].bid_size
            if bid_size == 0
            else Quantity.from_str(str(bid_size))
        )
        ask = self._last_quote_tick[instrument_id].ask if ask == 0 else Price.from_str(str(ask))
        ask_size = (
            self._last_quote_tick[instrument_id].ask_size
            if ask_size == 0
            else Quantity.from_str(str(ask_size))
        )

        quote_tick = QuoteTick(
            instrument_id=instrument_id,
            bid=bid,
            ask=ask,
            bid_size=bid_size,
            ask_size=ask_size,
            ts_event=self._clock.timestamp_ns(),
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(quote_tick)
        self._last_quote_tick[instrument_id] = quote_tick

    def _handle_24hr_ticker(self, data: List[Any], symbol: str):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        ticker: ZbTicker = parse_ticker_ws(
            instrument_id=instrument_id,
            msg=data,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(ticker)

    def _handle_trade(self, data: List[List[Any]], symbol: str, msg_type: Optional[str] = None):
        if msg_type and msg_type == "Whole":
            return

        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        for d in data:
            trade_tick: TradeTick = parse_trade_tick_ws(
                instrument_id=instrument_id,
                msg=d,
                ts_init=self._clock.timestamp_ns(),
            )
            self._handle_data(trade_tick)

    def _handle_kline(
        self,
        data: List[List[Any]],
        symbol: str,
        chan_type: str,
        msg_type: Optional[str] = None,
    ):
        if msg_type and msg_type == "Whole":
            return

        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        for d in data:
            last_kline = self._kline_buffer[instrument_id]
            self._kline_buffer[instrument_id] = d
            if not last_kline:
                continue
            elif last_kline[5] != d[5]:
                bar: ZbBar = parse_bar_ws(
                    instrument_id=instrument_id,
                    chan_type=chan_type,
                    kline=last_kline,
                    ts_init=self._clock.timestamp_ns(),
                )
                self._handle_data(bar)

    def _handle_mark_price(self, data: Union[Dict, str], symbol: str):
        if isinstance(data, dict):
            for sym, price in data.items():
                self._handle_single_mark_price(price, sym)
        else:
            self._handle_single_mark_price(data, symbol)

    def _handle_single_mark_price(self, price: str, symbol: str):
        instrument_id = self._instrument_provider.find_instrument_id_by_local_symbol(symbol)
        mark_price = parse_mark_price(
            instrument_id=instrument_id,
            price=price,
            ts_init=self._clock.timestamp_ns(),
        )
        self._handle_data(mark_price)

    def _handle_index_price(self, data: str, symbol: str):
        pass

    def _handle_funding_rate(self, data: Dict[str, Any], symbol: str):
        pass

    def _handle_mark_kline(
        self,
        data: List[List[Any]],
        symbol: str,
        chan_type: str,
        msg_type: Optional[str] = None,
    ):
        pass

    def _handle_index_kline(
        self,
        data: List[List[Any]],
        symbol: str,
        chan_type: str,
        msg_type: Optional[str] = None,
    ):
        pass



from typing import Dict

from nautilus_trader.model.enums import ContingencyType
from nautilus_trader.model.enums import LiquiditySide
from nautilus_trader.model.enums import OrderSide
from nautilus_trader.model.enums import OrderStatus
from nautilus_trader.model.enums import OrderType
from nautilus_trader.model.enums import PositionSide
from nautilus_trader.model.enums import TimeInForce
from nautilus_trader.model.events.account import AccountState
from nautilus_trader.model.events.order import OrderAccepted
from nautilus_trader.model.events.order import OrderCanceled
from nautilus_trader.model.events.order import OrderCancelRejected
from nautilus_trader.model.events.order import OrderDenied
from nautilus_trader.model.events.order import OrderEvent
from nautilus_trader.model.events.order import OrderExpired
from nautilus_trader.model.events.order import OrderFilled
from nautilus_trader.model.events.order import OrderInitialized
from nautilus_trader.model.events.order import OrderModifyRejected
from nautilus_trader.model.events.order import OrderPendingCancel
from nautilus_trader.model.events.order import OrderPendingUpdate
from nautilus_trader.model.events.order import OrderRejected
from nautilus_trader.model.events.order import OrderSubmitted
from nautilus_trader.model.events.order import OrderTriggered
from nautilus_trader.model.events.order import OrderUpdated
from nautilus_trader.model.events.position import PositionChanged
from nautilus_trader.model.events.position import PositionClosed
from nautilus_trader.model.events.position import PositionEvent
from nautilus_trader.model.events.position import PositionOpened
from nautilus_trader.model.orders.base import Order
from nautilus_trader.model.orders.limit import LimitOrder
from nautilus_trader.model.orders.market import MarketOrder
from nautilus_trader.model.orders.stop_limit import StopLimitOrder
from nautilus_trader.model.orders.stop_market import StopMarketOrder
from nautilus_trader.model.orders.trailing_stop_limit import TrailingStopLimitOrder
from nautilus_trader.model.orders.trailing_stop_market import TrailingStopMarketOrder
from nautilus_trader.model.position import Position
from schema_registry.client import schema


def enums_to_avro_type(cls) -> Dict:
    return {
        "type": "enum",
        "name": cls.__name__,
        "symbols": [e.name for e in cls],
    }


OrderEventSchema = schema.AvroSchema(
    {
        "type": "record",
        "name": OrderEvent.__name__,
        "namespace": "nautilus_trader.model.events.order",
        "fields": [
            {
                "name": "type",
                "type": {
                    "type": "enum",
                    "name": "OrderEventType",
                    "symbols": [cls.__name__ for cls in OrderEvent.__subclasses__()],
                },
            },
            {"name": "trader_id", "type": "string"},
            {"name": "strategy_id", "type": "string"},
            {"name": "instrument_id", "type": "string"},
            {"name": "client_order_id", "type": "string"},
            {"name": "venue_order_id", "type": ["string", "null"], "default": None},
            {"name": "trade_id", "type": ["string", "null"], "default": None},
            {"name": "position_id", "type": ["string", "null"], "default": None},
            {"name": "account_id", "type": ["string", "null"], "default": None},
            {
                "name": "order_side",
                "type": [enums_to_avro_type(OrderSide), "null"],
                "default": None,
            },
            {
                "name": "order_type",
                "type": [enums_to_avro_type(OrderType), "null"],
                "default": None,
            },
            {"name": "quantity", "type": ["float", "null"], "default": None},
            {
                "name": "time_in_force",
                "type": [enums_to_avro_type(TimeInForce), "null"],
                "default": None,
            },
            {"name": "post_only", "type": ["boolean", "null"], "default": None},
            {"name": "reduce_only", "type": ["boolean", "null"], "default": None},
            {"name": "options", "type": ["string", "null"], "default": None},
            {"name": "order_list_id", "type": ["string", "null"], "default": None},
            {
                "name": "contingency_type",
                "type": [enums_to_avro_type(ContingencyType), "null"],
                "default": None,
            },
            {"name": "linked_order_ids", "type": ["string", "null"], "default": None},
            {"name": "parent_order_id", "type": ["string", "null"], "default": None},
            {"name": "tags", "type": ["string", "null"], "default": None},
            {"name": "event_id", "type": "string"},
            {"name": "ts_event", "type": ["long", "null"], "default": None},
            {"name": "ts_init", "type": "long"},
            {"name": "reason", "type": ["string", "null"], "default": None},
            {"name": "reconciliation", "type": ["boolean", "null"], "default": None},
            {"name": "price", "type": ["float", "null"], "default": None},
            {"name": "trigger_price", "type": ["float", "null"], "default": None},
            {"name": "last_qty", "type": ["float", "null"], "default": None},
            {"name": "last_px", "type": ["float", "null"], "default": None},
            {"name": "currency", "type": ["string", "null"], "default": None},
            {"name": "commission", "type": ["string", "null"], "default": None},
            {"name": "liquidity_side", "type": ["string", "null"], "default": None},
            {"name": "info", "type": ["string", "null"], "default": None},
        ],
    }
)

PositionEventSchema = schema.AvroSchema(
    {
        "type": "record",
        "name": PositionEvent.__name__,
        "namespace": "nautilus_trader.model.events.position",
        "fields": [
            {
                "name": "type",
                "type": {
                    "type": "enum",
                    "name": "OrderEventType",
                    "symbols": [cls.__name__ for cls in PositionEvent.__subclasses__()],
                },
            },
            {"name": "trader_id", "type": "string"},
            {"name": "strategy_id", "type": "string"},
            {"name": "instrument_id", "type": "string"},
            {"name": "position_id", "type": "string"},
            {"name": "account_id", "type": "string"},
            {"name": "from_order", "type": "string"},
            {"name": "entry", "type": enums_to_avro_type(OrderSide)},
            {"name": "side", "type": enums_to_avro_type(PositionSide)},
            {"name": "net_qty", "type": "float"},
            {"name": "quantity", "type": "float"},
            {"name": "peak_qty", "type": "float"},
            {"name": "last_qty", "type": "float"},
            {"name": "last_px", "type": "float"},
            {"name": "currency", "type": "string"},
            {"name": "event_id", "type": "string"},
            {"name": "ts_init", "type": "long"},
            {"name": "ts_event", "type": ["long", "null"], "default": None},
            {"name": "ts_opened", "type": ["long", "null"], "default": None},
            {"name": "ts_closed", "type": ["long", "null"], "default": None},
            {"name": "duration_ns", "type": ["long", "null"], "default": None},
            {"name": "avg_px_open", "type": ["float", "null"], "default": None},
            {"name": "avg_px_close", "type": ["float", "null"], "default": None},
            {"name": "realized_return", "type": ["string", "null"], "default": None},
            {"name": "realized_pnl", "type": "string"},
        ],
    }
)

OrderSchema = schema.AvroSchema(
    {
        "type": "record",
        "name": Order.__name__,
        "namespace": "nautilus_trader.model.orders",
        "fields": [
            {"name": "trader_id", "type": "string"},
            {"name": "strategy_id", "type": "string"},
            {"name": "instrument_id", "type": "string"},
            {"name": "client_order_id", "type": "string"},
            {"name": "order_list_id", "type": ["string", "null"], "default": None},
            {"name": "position_id", "type": ["string", "null"], "default": None},
            {"name": "account_id", "type": ["string", "null"], "default": None},
            {"name": "last_trade_id", "type": ["string", "null"], "default": None},
            {"name": "type", "type": enums_to_avro_type(OrderType)},
            {"name": "side", "type": enums_to_avro_type(OrderSide)},
            {"name": "quantity", "type": "float"},
            {"name": "price", "type": ["float", "null"], "default": None},
            {"name": "time_in_force", "type": enums_to_avro_type(TimeInForce)},
            {"name": "is_reduce_only", "type": "boolean"},
            {"name": "filled_qty", "type": "float"},
            {"name": "avg_px", "type": ["float", "null"], "default": None},
            {"name": "slippage", "type": "float"},
            {"name": "status", "type": enums_to_avro_type(OrderStatus)},
            {"name": "contingency_type", "type": enums_to_avro_type(ContingencyType)},
            {"name": "linked_order_ids", "type": ["string", "null"], "default": None},
            {"name": "parent_order_id", "type": ["string", "null"], "default": None},
            {"name": "tags", "type": ["string", "null"], "default": None},
            {"name": "ts_last", "type": "long"},
            {"name": "ts_init", "type": "long"},
            {"name": "expire_time_ns", "type": ["long", "null"], "default": None},
            {"name": "liquidity_side", "type": enums_to_avro_type(LiquiditySide)},
            {"name": "is_post_only", "type": "boolean"},
            {"name": "display_qty", "type": ["float", "null"], "default": None},
            {"name": "trigger_price", "type": ["float", "null"], "default": None},
            {"name": "trigger_type", "type": ["string", "null"], "default": None},
            {"name": "limit_offset", "type": ["string", "null"], "default": None},
            {"name": "trailing_offset", "type": ["string", "null"], "default": None},
            {"name": "offset_type", "type": ["string", "null"], "default": None},
        ],
    }
)


NAUTILUS_OBJECT_AVRO_SCHEMA = {
    AccountState.__name__: schema.AvroSchema(
        {
            "type": "record",
            "name": AccountState.__name__,
            "namespace": "nautilus_trader.model.events",
            "fields": [
                {"name": "account_id", "type": "string"},
                {"name": "account_type", "type": "string"},
                {"name": "base_currency", "type": ["string", "null"], "default": None},
                {"name": "balances", "type": "string"},
                {"name": "positions", "type": "string"},
                {"name": "reported", "type": "boolean"},
                {"name": "event_id", "type": "string"},
                {"name": "ts_event", "type": "long"},
                {"name": "ts_init", "type": "long"},
                {"name": "equity", "type": "float"},
                {"name": "equities", "type": "string", "default": "[]"},
            ],
        }
    ),
    # OrderEvent
    OrderAccepted.__name__: OrderEventSchema,
    OrderCancelRejected.__name__: OrderEventSchema,
    OrderCanceled.__name__: OrderEventSchema,
    OrderDenied.__name__: OrderEventSchema,
    OrderExpired.__name__: OrderEventSchema,
    OrderFilled.__name__: OrderEventSchema,
    OrderInitialized.__name__: OrderEventSchema,
    OrderPendingCancel.__name__: OrderEventSchema,
    OrderPendingUpdate.__name__: OrderEventSchema,
    OrderRejected.__name__: OrderEventSchema,
    OrderSubmitted.__name__: OrderEventSchema,
    OrderTriggered.__name__: OrderEventSchema,
    OrderModifyRejected.__name__: OrderEventSchema,
    OrderUpdated.__name__: OrderEventSchema,
    # PositionEvent
    PositionOpened.__name__: PositionEventSchema,
    PositionChanged.__name__: PositionEventSchema,
    PositionClosed.__name__: PositionEventSchema,
    # Orders
    LimitOrder.__name__: OrderSchema,
    MarketOrder.__name__: OrderSchema,
    StopLimitOrder.__name__: OrderSchema,
    StopMarketOrder.__name__: OrderSchema,
    TrailingStopLimitOrder.__name__: OrderSchema,
    TrailingStopMarketOrder.__name__: OrderSchema,
    # Position
    Position.__name__: schema.AvroSchema(
        {
            "type": "record",
            "name": Position.__name__,
            "namespace": "nautilus_trader.model.position",
            "fields": [
                {"name": "position_id", "type": "string"},
                {"name": "account_id", "type": "string"},
                {"name": "from_order", "type": "string"},
                {"name": "strategy_id", "type": "string"},
                {"name": "instrument_id", "type": "string"},
                {"name": "entry", "type": enums_to_avro_type(OrderSide)},
                {"name": "side", "type": enums_to_avro_type(PositionSide)},
                {"name": "net_qty", "type": "float"},
                {"name": "quantity", "type": "float"},
                {"name": "peak_qty", "type": "float"},
                {"name": "ts_opened", "type": "long"},
                {"name": "ts_closed", "type": "long"},
                {"name": "duration_ns", "type": "long"},
                {"name": "avg_px_open", "type": "float"},
                {"name": "avg_px_close", "type": ["float", "null"], "default": None},
                {"name": "quote_currency", "type": "string"},
                {"name": "base_currency", "type": "string"},
                {"name": "cost_currency", "type": "string"},
                {"name": "realized_return", "type": "string"},
                {"name": "realized_pnl", "type": "string"},
                {"name": "commissions", "type": "string"},
            ],
        }
    ),
}



from typing import Any, Dict, List, Optional

from nautilus_trader.common.logging import Logger
from nautilus_trader.common.providers import InstrumentProvider
from nautilus_trader.config import InstrumentProviderConfig
from nautilus_trader.core.correctness import PyCondition
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import Venue

from nacre.adapters.zb.common import format_websocket_market
from nacre.adapters.zb.common.enums import ZbAccountType
from nacre.adapters.zb.http.api.future_market import ZbFutureMarketHttpAPI
from nacre.adapters.zb.http.api.spot_market import ZbSpotMarketHttpAPI
from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.parsing import parse_future_instrument_http
from nacre.adapters.zb.parsing import parse_spot_instrument_http


class ZbInstrumentProvider(InstrumentProvider):
    """
    An example template of an ``InstrumentProvider`` showing the minimal methods
    which must be implemented for an integration to be complete.
    """

    def __init__(
        self,
        logger: Logger,
        venue: Venue,
        client: ZbHttpClient,
        account_type: ZbAccountType,
        config: Optional[InstrumentProviderConfig] = None,
    ):
        super().__init__(
            venue=venue,
            logger=logger,
            config=config,
        )

        self._client = client
        self._account_type = account_type

        if account_type.is_spot:
            self._market = ZbSpotMarketHttpAPI(self._client)
        elif account_type.is_futures:
            self._market = ZbFutureMarketHttpAPI(self._client)
        else:
            raise ValueError(f"{account_type} not support by zb for now")

        self._local_symbols_to_instrument_id: Dict[str, InstrumentId] = {}
        self._local_market_id_to_instrument_id: Dict[str, InstrumentId] = {}

    async def load_all_async(self, filters: Optional[Dict] = None) -> None:
        """
        Load the latest ZB instruments into the provider asynchronously.
        """
        if self._account_type.is_futures:
            await self.load_future_markets()
        elif self._account_type.is_spot:
            await self.load_spot_markets()

    async def load_ids_async(
        self,
        instrument_ids: List[InstrumentId],
        filters: Optional[Dict] = None,
    ) -> None:
        if not instrument_ids:
            self._log.info("No instrument IDs given for loading.")
            return

        # Check all instrument IDs
        for instrument_id in instrument_ids:
            PyCondition.equal(instrument_id.venue, self.venue, "instrument_id.venue", "self.venue")

        filters_str = "..." if not filters else f" with filters {filters}..."
        self._log.info(f"Loading instruments {instrument_ids}{filters_str}.")

        if self._account_type.is_futures:
            await self.load_future_markets(instrument_ids)
        elif self._account_type.is_spot:
            await self.load_spot_markets(instrument_ids)

    async def load_future_markets(self, instrument_ids: Optional[List[InstrumentId]] = None):
        markets: List[Dict] = await self._market.market_list()
        for market in markets:
            instrument = parse_future_instrument_http(market, self.venue)
            if instrument_ids is not None and instrument.id not in instrument_ids:
                continue

            self.add_currency(currency=instrument.base_currency)
            self.add_currency(currency=instrument.quote_currency)
            self.add(instrument=instrument)

            self._local_symbols_to_instrument_id[market["symbol"]] = instrument.id
            self._local_market_id_to_instrument_id[market["id"]] = instrument.id

    async def load_spot_markets(self, instrument_ids: Optional[List[InstrumentId]] = None):
        markets: Dict[str, Any] = await self._market.markets()
        for market, entry in markets.items():
            instrument = parse_spot_instrument_http(market, entry, self.venue)
            if instrument_ids is not None and instrument.id not in instrument_ids:
                continue

            self.add_currency(currency=instrument.base_currency)
            self.add_currency(currency=instrument.quote_currency)
            self.add(instrument=instrument)

            self._local_symbols_to_instrument_id[market] = instrument.id
            # Spot websocket return symbol like "BTC/USDT" in "btcusdt", map it here
            ws_symbol = format_websocket_market(market)
            self._local_market_id_to_instrument_id[ws_symbol] = instrument.id

    def find_instrument_id_by_local_symbol(self, local_symbol: str) -> InstrumentId:
        return self._local_symbols_to_instrument_id[local_symbol]

    def find_instrument_id_by_local_market_id(self, market_id: str) -> InstrumentId:
        return self._local_market_id_to_instrument_id[market_id]

    def list_local_market_ids(self) -> List[str]:
        return list(self._local_market_id_to_instrument_id.keys())



from typing import List

from nautilus_trader.config import ActorConfig


class ExposerConfig(ActorConfig):
    """
    Configuration for ``PubSub`` instances.

    host : str
        listen interface
    port : int
        listen port
    """

    host: str = "127.0.0.1"
    port: int = 8080


class PubSubConfig(ActorConfig):
    """
    Configuration for ``PubSub`` instances.

    topic_filters : list
        List of topic filters separated by comma.
    """

    topic_filters: List[str] = []


class KafkaConfig(PubSubConfig):
    """
    Configuration for ``KafkaPubSub`` instances.

    bootstrap_servers : str
        Kafka bootstrap_servers
    security_protocol : str
        Security protocol
    schema_registry : bool
        Confluentinc schema registry url.
    """

    bootstrap_servers: str = "localhost:9092"
    security_protocol: str = "PLAINTEXT"
    schema_registry: str = "http://localhost:8081"



from typing import Any, Dict, Optional

from nautilus_trader.core.correctness import PyCondition

from nacre.adapters.zb.common import format_market
from nacre.adapters.zb.http.spot import ZbSpotHttpClient


class ZbSpotMarketHttpAPI:
    """
    Provides access to the `ZB FUTURE Market` HTTP REST API.
    """

    BASE_ENDPOINT = "/data/v1/"

    def __init__(self, client: ZbSpotHttpClient):
        """
        Initialize a new instance of the ``ZbSpotMarketHttpAPI`` class.

        Parameters
        ----------
        client : ZbSpotHttpClient
            The Binance REST API client.

        """
        PyCondition.not_none(client, "client")

        self.client = client

    async def markets(self) -> Dict[str, Any]:
        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "markets",
        )

    async def all_ticker(self) -> Dict[str, Any]:
        return await self.client.query(url_path=self.BASE_ENDPOINT + "allTicker")

    async def ticker(self, market: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {"market": format_market(market)}

        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "ticker",
            payload=payload,
        )

    async def depth(
        self, market: str, size: Optional[int] = None, merge: Optional[float] = None
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"market": format_market(market)}
        if size is not None:
            payload["size"] = size
        if merge is not None:
            payload["merge"] = merge

        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "depth",
            payload=payload,
        )

    async def trades(self, market: str, since: Optional[int] = None) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"market": format_market(market)}
        if since is not None:
            payload["since"] = since

        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "trades",
            payload=payload,
        )

    async def kline(
        self,
        market: str,
        type: Optional[str] = None,
        since: Optional[int] = None,
        size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"market": format_market(market)}
        if type is not None:
            payload["type"] = type
        if since is not None:
            payload["since"] = since
        if size is not None:
            payload["size"] = size

        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "kline",
            payload=payload,
        )

    async def group_markets(self) -> Dict[str, Any]:
        return await self.client.query(
            url_path=self.BASE_ENDPOINT + "getGroupMarkets",
        )



from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.core.datetime cimport nanos_to_millis
from nautilus_trader.model.events.order cimport OrderAccepted
from nautilus_trader.model.events.order cimport OrderCanceled
from nautilus_trader.model.events.order cimport OrderCancelRejected
from nautilus_trader.model.events.order cimport OrderDenied
from nautilus_trader.model.events.order cimport OrderEvent
from nautilus_trader.model.events.order cimport OrderExpired
from nautilus_trader.model.events.order cimport OrderFilled
from nautilus_trader.model.events.order cimport OrderInitialized
from nautilus_trader.model.events.order cimport OrderModifyRejected
from nautilus_trader.model.events.order cimport OrderPendingCancel
from nautilus_trader.model.events.order cimport OrderPendingUpdate
from nautilus_trader.model.events.order cimport OrderRejected
from nautilus_trader.model.events.order cimport OrderSubmitted
from nautilus_trader.model.events.order cimport OrderTriggered
from nautilus_trader.model.events.order cimport OrderUpdated
from nautilus_trader.model.events.position cimport PositionChanged
from nautilus_trader.model.events.position cimport PositionClosed
from nautilus_trader.model.events.position cimport PositionEvent
from nautilus_trader.model.events.position cimport PositionOpened
from nautilus_trader.model.orders.base cimport Order
from nautilus_trader.model.orders.limit cimport LimitOrder
from nautilus_trader.model.orders.market cimport MarketOrder
from nautilus_trader.model.orders.stop_limit cimport StopLimitOrder
from nautilus_trader.model.orders.stop_market cimport StopMarketOrder
from nautilus_trader.model.orders.trailing_stop_limit cimport TrailingStopLimitOrder
from nautilus_trader.model.orders.trailing_stop_market cimport TrailingStopMarketOrder
from nautilus_trader.model.position cimport Position
from nautilus_trader.serialization.base cimport _OBJECT_TO_DICT_MAP

from schema_registry.client import AsyncSchemaRegistryClient
from schema_registry.serializers import AsyncAvroMessageSerializer

from nacre.serialization.base cimport AsyncSerializer

from nacre.serialization.avro.schema import NAUTILUS_OBJECT_AVRO_SCHEMA


_AVRO_TO_DICT_MAP = {
    # Orders
    LimitOrder.__name__: lambda obj: obj.to_dict(),
    MarketOrder.__name__: lambda obj: obj.to_dict(),
    StopLimitOrder.__name__: lambda obj: obj.to_dict(),
    StopMarketOrder.__name__: lambda obj: obj.to_dict(),
    TrailingStopLimitOrder.__name__: lambda obj: obj.to_dict(),
    TrailingStopMarketOrder.__name__: lambda obj: obj.to_dict(),

    # Position
    Position.__name__: lambda obj: obj.to_dict(),
}


# Kafka postgresql connector require "-value", "-key" suffix on topic
_AVRO_TO_SUBJECT_MAP = {

    # Orders
    LimitOrder.__name__: "states.order-value",
    MarketOrder.__name__: "states.order-value",
    StopLimitOrder.__name__: "states.order-value",
    StopMarketOrder.__name__: "states.order-value",
    TrailingStopLimitOrder.__name__: "states.order-value",
    TrailingStopMarketOrder.__name__: "states.order-value",

    OrderAccepted.__name__: "events.order-value",
    OrderCancelRejected.__name__: "events.order-value",
    OrderCanceled.__name__: "events.order-value",
    OrderDenied.__name__: "events.order-value",
    OrderExpired.__name__: "events.order-value",
    OrderFilled.__name__: "events.order-value",
    OrderInitialized.__name__: "events.order-value",
    OrderPendingCancel.__name__: "events.order-value",
    OrderPendingUpdate.__name__: "events.order-value",
    OrderRejected.__name__: "events.order-value",
    OrderSubmitted.__name__: "events.order-value",
    OrderTriggered.__name__: "events.order-value",
    OrderModifyRejected.__name__: "events.order-value",
    OrderUpdated.__name__: "events.order-value",

    # PositionEvent
    PositionOpened.__name__: "events.position-value",
    PositionChanged.__name__: "events.position-value",
    PositionClosed.__name__: "events.position-value",

    # Position
    Position.__name__: "states.position-value",
}


cdef class AvroSerializer(AsyncSerializer):
    def __init__(self, schema_registry: str):
        self._client = AsyncSchemaRegistryClient(schema_registry)
        self._async_serializer = AsyncAvroMessageSerializer(self._client)

    async def serialize(self, object obj) -> bytes:
        Condition.not_none(obj, "obj")

        cls_name = type(obj).__name__

        delegate = _AVRO_TO_DICT_MAP.get(cls_name)
        if delegate is None:
            delegate = _OBJECT_TO_DICT_MAP.get(cls_name)
        if delegate is None:
            raise TypeError(f"Cannot serialize object `{cls_name}`.")

        obj_dict = delegate(obj)
        schema = NAUTILUS_OBJECT_AVRO_SCHEMA.get(cls_name)
        if schema is None:
            raise TypeError(f"Cannot get schema for class `{cls_name}`")

        subject = _AVRO_TO_SUBJECT_MAP.get(cls_name)
        if subject is None:
            raise TypeError(f"Cannot get subject for class `{cls_name}`")

        obj_dict = self.type_convert(obj, obj_dict)
        return await self._async_serializer.encode_record_with_schema(subject, schema, obj_dict)

    async def deserialize(self, obj_bytes: bytes) -> object:
        pass

    async def stop(self):
        await self._client.session.aclose()

    cpdef dict type_convert(self, object obj, dict obj_dict):
        # Convert timestamp fields to millisecond
        cdef str field_name
        for field_name in ("ts_event", "ts_init", "ts_last", "ts_opened", "ts_closed"):
            field = obj_dict.get(field_name)
            if field is not None:
                obj_dict[field_name] = nanos_to_millis(field)

        if isinstance(obj, (Order, OrderEvent)):
            return self.order_type_convert(obj, obj_dict)
        elif isinstance(obj, (Position, PositionEvent)):
            return self.position_type_convert(obj, obj_dict)
        return obj_dict

    cpdef dict order_type_convert(self, object obj, dict obj_dict):
        # Convert str typed value back to float
        cdef str field_name
        for field_name in ("quantity", "price", "filled_qty", "avg_px", "slippage", "display_qty", "trigger_price", "last_qty", "last_px"):
            field = obj_dict.get(field_name)
            if field is not None:
                obj_dict[field_name] = float(field)

        # Fix MarketOrder reduce_only name inconsistence
        reduce_flag = obj_dict.get("reduce_only")
        if reduce_flag is not None:
            obj_dict["is_reduce_only"] = reduce_flag

        return obj_dict

    cpdef dict position_type_convert(self, object obj, dict obj_dict):
        # Convert str typed value back to float
        cdef str field_name
        for field_name in ("net_qty", "quantity", "peak_qty", "avg_px_open", "realized_points", "last_qty", "last_px"):
            field = obj_dict.get(field_name)
            if field is not None:
                obj_dict[field_name] = float(field)
        # Bug here
        avg_px_close = obj_dict.get("avg_px_close")
        if avg_px_close is not None:
            if avg_px_close == "None":
                obj_dict["avg_px_close"] = None
            else:
                obj_dict["avg_px_close"] = float(avg_px_close)
        return obj_dict
